/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm8_channel.h"
#include "uvm8_global.h"
#include "uvm8_hal.h"
#include "uvm8_pmm_gpu.h"
#include "uvm8_push.h"
#include "uvm8_test.h"
#include "uvm8_tracker.h"
#include "uvm8_va_space.h"

#define CE_TEST_MEM_SIZE (2 * 1024 * 1024)
#define CE_TEST_MEM_END_SIZE 32
#define CE_TEST_MEM_BEGIN_SIZE 32
#define CE_TEST_MEM_MIDDLE_SIZE (CE_TEST_MEM_SIZE - CE_TEST_MEM_BEGIN_SIZE - CE_TEST_MEM_END_SIZE)
#define CE_TEST_MEM_MIDDLE_OFFSET (CE_TEST_MEM_BEGIN_SIZE)
#define CE_TEST_MEM_END_OFFSET (CE_TEST_MEM_SIZE - CE_TEST_MEM_BEGIN_SIZE)
#define CE_TEST_MEM_COUNT 5

static NV_STATUS test_non_pipelined(uvm_gpu_t *gpu)
{
    NvU32 i;
    NV_STATUS status;
    uvm_rm_mem_t *mem[CE_TEST_MEM_COUNT] = { NULL };
    uvm_rm_mem_t *host_mem = NULL;
    NvU32 *host_ptr;
    NvU64 host_mem_gpu_va;
    NvU64 dst_va;
    NvU64 src_va;
    uvm_push_t push;

    status = uvm_rm_mem_alloc_and_map_cpu(gpu, UVM_RM_MEM_TYPE_SYS, CE_TEST_MEM_SIZE, &host_mem);
    TEST_CHECK_GOTO(status == NV_OK, done);
    host_ptr = (NvU32 *)uvm_rm_mem_get_cpu_va(host_mem);
    host_mem_gpu_va = uvm_rm_mem_get_gpu_va(host_mem, gpu);
    memset(host_ptr, 0, CE_TEST_MEM_SIZE);

    for (i = 0; i < CE_TEST_MEM_COUNT; ++i) {
        status = uvm_rm_mem_alloc(gpu, UVM_RM_MEM_TYPE_GPU, CE_TEST_MEM_SIZE, &mem[i]);
        TEST_CHECK_GOTO(status == NV_OK, done);
    }

    status = uvm_push_begin(gpu->channel_manager, UVM_CHANNEL_TYPE_GPU_INTERNAL, &push, "Non-pipelined test");
    TEST_CHECK_GOTO(status == NV_OK, done);

    // All of the following CE transfers are done from a single (L)CE and
    // disabling pipelining is enough to order them when needed. Only push_end
    // needs a MEMBAR SYS to order everything with the CPU.

    // Initialize to a bad value
    for (i = 0; i < CE_TEST_MEM_COUNT; ++i) {
        uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_PIPELINED);
        uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_MEMBAR_NONE);
        gpu->ce_hal->memset_v_4(&push, uvm_rm_mem_get_gpu_va(mem[i], gpu), 1337 + i, CE_TEST_MEM_SIZE);
    }

    // Set the first buffer to 1
    uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_MEMBAR_NONE);
    gpu->ce_hal->memset_v_4(&push, uvm_rm_mem_get_gpu_va(mem[0], gpu), 1, CE_TEST_MEM_SIZE);

    for (i = 0; i < CE_TEST_MEM_COUNT; ++i) {
        NvU32 dst = i + 1;
        if (dst == CE_TEST_MEM_COUNT)
            dst_va = host_mem_gpu_va;
        else
            dst_va = uvm_rm_mem_get_gpu_va(mem[dst], gpu);

        src_va = uvm_rm_mem_get_gpu_va(mem[i], gpu);

        // The first memcpy needs to be non-pipelined as otherwise the previous
        // memset/memcpy to the source may not be done yet.

        // Alternate the order of copying the beginning and the end
        if (i % 2 == 0) {
            uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_MEMBAR_NONE);
            gpu->ce_hal->memcopy_v_to_v(&push, dst_va + CE_TEST_MEM_END_OFFSET, src_va + CE_TEST_MEM_END_OFFSET, CE_TEST_MEM_END_SIZE);

            uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_MEMBAR_NONE);
            uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_PIPELINED);
            gpu->ce_hal->memcopy_v_to_v(&push, dst_va + CE_TEST_MEM_MIDDLE_OFFSET, src_va + CE_TEST_MEM_MIDDLE_OFFSET, CE_TEST_MEM_MIDDLE_SIZE);

            uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_MEMBAR_NONE);
            uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_PIPELINED);
            gpu->ce_hal->memcopy_v_to_v(&push, dst_va, src_va, CE_TEST_MEM_BEGIN_SIZE);
        }
        else {
            uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_MEMBAR_NONE);
            gpu->ce_hal->memcopy_v_to_v(&push, dst_va, src_va, CE_TEST_MEM_BEGIN_SIZE);

            uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_MEMBAR_NONE);
            uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_PIPELINED);
            gpu->ce_hal->memcopy_v_to_v(&push, dst_va + CE_TEST_MEM_MIDDLE_OFFSET, src_va + CE_TEST_MEM_MIDDLE_OFFSET, CE_TEST_MEM_MIDDLE_SIZE);

            uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_MEMBAR_NONE);
            uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_PIPELINED);
            gpu->ce_hal->memcopy_v_to_v(&push, dst_va + CE_TEST_MEM_END_OFFSET, src_va + CE_TEST_MEM_END_OFFSET, CE_TEST_MEM_END_SIZE);
        }
    }

    status = uvm_push_end_and_wait(&push);
    TEST_CHECK_GOTO(status == NV_OK, done);


    for (i = 0; i < CE_TEST_MEM_SIZE / sizeof(NvU32); ++i) {
        if (host_ptr[i] != 1) {
            UVM_TEST_PRINT("host_ptr[%u] = %u instead of 1\n", i, host_ptr[i]);
            status = NV_ERR_INVALID_STATE;
            goto done;
        }
    }

done:
    for (i = 0; i < CE_TEST_MEM_COUNT; ++i) {
        uvm_rm_mem_free(mem[i]);
    }
    uvm_rm_mem_free(host_mem);

    return status;
}

#define REDUCTIONS 32

static NV_STATUS test_membar(uvm_gpu_t *gpu)
{
    NvU32 i;
    NV_STATUS status;
    uvm_rm_mem_t *host_mem = NULL;
    NvU32 *host_ptr;
    NvU64 host_mem_gpu_va;
    uvm_push_t push;
    NvU32 value;

    if (!uvm_gpu_is_gk110_plus(gpu)) {
        // semaphore_reduction_inc is only supported on GK110+
        return NV_OK;
    }

    status = uvm_rm_mem_alloc_and_map_cpu(gpu, UVM_RM_MEM_TYPE_SYS, sizeof(NvU32), &host_mem);
    TEST_CHECK_GOTO(status == NV_OK, done);
    host_ptr = (NvU32 *)uvm_rm_mem_get_cpu_va(host_mem);
    host_mem_gpu_va = uvm_rm_mem_get_gpu_va(host_mem, gpu);
    *host_ptr = 0;

    status = uvm_push_begin(gpu->channel_manager, UVM_CHANNEL_TYPE_GPU_TO_CPU, &push, "Membar test");
    TEST_CHECK_GOTO(status == NV_OK, done);

    for (i = 0; i < REDUCTIONS; ++i) {
        uvm_push_set_flag(&push, UVM_PUSH_FLAG_CE_NEXT_MEMBAR_NONE);
        gpu->ce_hal->semaphore_reduction_inc(&push, host_mem_gpu_va, REDUCTIONS + 1);
    }

    // Without a sys membar the channel tracking semaphore can and does complete
    // before all the reductions.
    status = uvm_push_end_and_wait(&push);
    TEST_CHECK_GOTO(status == NV_OK, done);

    value = *host_ptr;
    if (value != REDUCTIONS) {
        UVM_TEST_PRINT("Value = %u instead of %u, GPU %s\n", value, REDUCTIONS, gpu->name);
        status = NV_ERR_INVALID_STATE;
        goto done;
    }

done:
    uvm_rm_mem_free(host_mem);

    return status;
}

#define MEM_TEST_SIZE UVM_CHUNK_SIZE_64K
#define MEM_TEST_SIZE_HALF (MEM_TEST_SIZE / 2)
#define MEM_TEST_ITERS 4

static NV_STATUS test_memcpy_and_memset(uvm_gpu_t *gpu)
{
    NV_STATUS status;
    struct page *sys_phys_pages_verif = NULL;
    struct page *sys_phys_pages = NULL;
    uvm_gpu_chunk_t *gpu_phys_chunk = NULL;
    uvm_rm_mem_t *gpu_virt = NULL;
    uvm_rm_mem_t *host_virt = NULL;
    uvm_gpu_address_t test_mem[4];
    uvm_gpu_address_t verif_addr;
    NvU32 *verif32;
    NvU64 *verif64;
    int i, j, k, l;

    sys_phys_pages_verif = alloc_pages(NV_UVM_GFP_FLAGS | __GFP_ZERO, get_order(MEM_TEST_SIZE));
    if (!sys_phys_pages_verif) {
        status = NV_ERR_NO_MEMORY;
        goto done;
    }
    verif_addr = uvm_gpu_address_physical(UVM_APERTURE_SYS, page_to_phys(sys_phys_pages_verif));
    verif32 = (NvU32*)phys_to_virt(verif_addr.address);
    verif64 = (NvU64*)verif32;

    sys_phys_pages = alloc_pages(NV_UVM_GFP_FLAGS | __GFP_ZERO, get_order(MEM_TEST_SIZE));
    if (!sys_phys_pages) {
        status = NV_ERR_NO_MEMORY;
        goto done;
    }
    test_mem[0] = uvm_gpu_address_physical(UVM_APERTURE_SYS, page_to_phys(sys_phys_pages));

    status = uvm_pmm_gpu_alloc(&gpu->pmm, 1, MEM_TEST_SIZE, UVM_PMM_GPU_MEMORY_TYPE_USER, &gpu_phys_chunk, UVM_PMM_ALLOC_FLAGS_NONE);
    if (status != NV_OK)
        goto done;
    test_mem[1] = uvm_gpu_address_physical(UVM_APERTURE_VID, gpu_phys_chunk->address);

    status = uvm_rm_mem_alloc(gpu, UVM_RM_MEM_TYPE_GPU, MEM_TEST_SIZE, &gpu_virt);
    if (status != NV_OK)
        goto done;
    test_mem[2] = uvm_gpu_address_virtual(uvm_rm_mem_get_gpu_va(gpu_virt, gpu));

    status = uvm_rm_mem_alloc(gpu, UVM_RM_MEM_TYPE_SYS, MEM_TEST_SIZE, &host_virt);
    if (status != NV_OK)
        goto done;
    test_mem[3] = uvm_gpu_address_virtual(uvm_rm_mem_get_gpu_va(host_virt, gpu));

    for (i = 0; i < MEM_TEST_ITERS; ++i) {
        for (j = 0; j < ARRAY_SIZE(test_mem); ++j) {
            for (k = 0; k < ARRAY_SIZE(test_mem); ++k) {
                uvm_push_t push;
                uvm_gpu_address_t src = test_mem[j];
                uvm_gpu_address_t dst = test_mem[k];
                NvU64 value64 = (i + 2) * (1ull << 32) + (i + 1);
                NvU32 value32 = (NvU32)value64;

                const char *src_type = src.virtual ? "virtual" : "physical";
                const char *src_loc = src.aperture == UVM_APERTURE_SYS ? "sysmem" : "vidmem";
                const char *dst_type = dst.virtual ? "virtual" : "physical";
                const char *dst_loc = dst.aperture == UVM_APERTURE_SYS ? "sysmem" : "vidmem";

                uvm_gpu_address_t src_mid = src;
                src_mid.address += MEM_TEST_SIZE_HALF;

                status = uvm_push_begin(gpu->channel_manager, UVM_CHANNEL_TYPE_GPU_INTERNAL, &push,
                        "Memset %s %s (0x%llx) and memcopy to %s %s (0x%llx), iter %d",
                        src_type, src_loc, src.address, dst_type, dst_loc, dst.address, i);
                TEST_CHECK_GOTO(status == NV_OK, done);

                // Memset first half with 4-byte value and second half with
                // 8-byte value, then memcpy both to dst and from dst to the
                // verif location (physical sysmem).
                gpu->ce_hal->memset_4(&push, src, value32, MEM_TEST_SIZE_HALF);
                gpu->ce_hal->memset_8(&push, src_mid, value64, MEM_TEST_SIZE_HALF);
                gpu->ce_hal->memcopy(&push, dst, src, MEM_TEST_SIZE);
                gpu->ce_hal->memcopy(&push, verif_addr, dst, MEM_TEST_SIZE);

                status = uvm_push_end_and_wait(&push);
                TEST_CHECK_GOTO(status == NV_OK, done);

                for (l = 0; l < MEM_TEST_SIZE_HALF / sizeof(NvU32); ++l) {
                    if (verif32[l] != value32) {
                        UVM_TEST_PRINT("Memset 4 of %s %s and memcpy into %s %s failed, value32[%d] = 0x%x instead of 0x%x\n",
                                src_type, src_loc, dst_type, dst_loc,
                                l, verif32[l], value32);
                        status = NV_ERR_INVALID_STATE;
                        goto done;
                    }
                }
                for (l = MEM_TEST_SIZE_HALF / sizeof(NvU64); l < MEM_TEST_SIZE / sizeof(NvU64); ++l) {
                    if (verif64[l] != value64) {
                        UVM_TEST_PRINT("Memset 8 of %s %s and memcpy into %s %s failed, verif64[%d] = 0x%llx instead of 0x%llx\n",
                                src_type, src_loc, dst_type, dst_loc,
                                l, verif64[l], value64);
                        status = NV_ERR_INVALID_STATE;
                        goto done;
                    }
                }
            }
        }
    }

done:
    if (host_virt)
        uvm_rm_mem_free(host_virt);
    if (gpu_virt)
        uvm_rm_mem_free(gpu_virt);
    if (gpu_phys_chunk) {
        uvm_pmm_gpu_lock_chunk(&gpu->pmm, gpu_phys_chunk);
        uvm_pmm_gpu_free(&gpu->pmm, MEM_TEST_SIZE, UVM_PMM_GPU_MEMORY_TYPE_USER, gpu_phys_chunk);
    }
    if (sys_phys_pages)
        __free_pages(sys_phys_pages, get_order(MEM_TEST_SIZE));
    if (sys_phys_pages_verif)
        __free_pages(sys_phys_pages_verif, get_order(MEM_TEST_SIZE));

    return status;
}

static NV_STATUS test_ce(uvm_va_space_t *va_space)
{
    uvm_gpu_t *gpu;

    for_each_global_gpu(gpu) {
        TEST_CHECK_RET(test_non_pipelined(gpu) == NV_OK);
        TEST_CHECK_RET(test_membar(gpu) == NV_OK);
        TEST_CHECK_RET(test_memcpy_and_memset(gpu) == NV_OK);
    }

    return NV_OK;
}

NV_STATUS uvm8_test_ce_sanity(UVM_TEST_CE_SANITY_PARAMS *params, struct file *filp)
{
    NV_STATUS status;
    uvm_va_space_t *va_space = uvm_va_space_get(filp);

    uvm_down_read(&g_uvm_global.global_lock);
    uvm_down_read(&va_space->lock);

    status = test_ce(va_space);
    if (status != NV_OK)
        goto done;

done:
    uvm_up_read(&va_space->lock);
    uvm_up_read(&g_uvm_global.global_lock);

    return status;
}
