/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#ifndef __UVM8_GPU_H__
#define __UVM8_GPU_H__

#include "nvtypes.h"
#include "nvmisc.h"
#include "uvmtypes.h"
#include "uvm_linux.h"
#include "uvm_common.h"
#include "ctrl2080mc.h"
#include "uvm8_forward_decl.h"
#include "uvm8_pmm_gpu.h"

#include <linux/bitmap.h>

// TODO: Replace by NV0080_MAX_DEVICES
#define UVM8_MAX_GPUS 32
#define UVM8_MAX_PROCESSORS (UVM8_MAX_GPUS + 1)
#define UVM8_MAX_UNIQUE_GPU_PAIRS SUM_FROM_0_TO_N(UVM8_MAX_GPUS - 1)

// Static processor id for the CPU
#define UVM_CPU_ID (0)

// A unique uvm internal gpu id, guaranteed to be in the [1, UVM8_MAX_PROCESSORS) range
typedef NvU32 uvm_gpu_id_t;

// Helper to iterate over all valid gpu ids
#define for_each_gpu_id(i) for (i = 1; i < UVM8_MAX_PROCESSORS; ++i)

// Same as gpu id type but can also include the cpu id (UVM_CPU_ID)
typedef NvU32 uvm_processor_id_t;

// A collection of uvm processor ids
// Operated on with the uvm_processor_mask_*() functions below
typedef struct
{
    DECLARE_BITMAP(bitmap, UVM8_MAX_PROCESSORS);
} uvm_processor_mask_t;

// Buffer length to store uvm gpu id, RM device name and gpu uuid.
#define UVM_GPU_NICE_NAME_BUFFER_LENGTH (sizeof("ID 999: : ") + \
            UVM_GPU_NAME_LENGTH + UVM_GPU_UUID_TEXT_BUFFER_LENGTH)

struct uvm_gpu_struct
{
    // Reference count for how many places are holding onto a GPU (internal to UVM driver).
    // This includes any GPUs we know about, not just GPUs that are registered with a VA space.
    // Most GPUs end up being registered, but there are brief periods when they are not
    // registered, such as during interrupt handling, and in add_gpu() or remove_gpu().
    struct kref gpu_kref;

    // Refcount of the gpu, i.e. how many times it has been retained. This means that it is
    // also a count of how many times it has been registered with a VA space.
    //
    // Security note: keep it as a 64bit counter to prevent overflow cases
    // (a user can create a lot of va spaces and register the gpu with them).
    NvU64 registered_ref_count;

    // A unique uvm gpu id in range [1, UVM8_MAX_PROCESSORS)
    uvm_gpu_id_t id;

    // The gpu's uuid
    NvProcessorUuid uuid;

    // Nice printable name including the uvm gpu id, ascii name from RM and uuid
    char name[UVM_GPU_NICE_NAME_BUFFER_LENGTH];

    // Gpu architecture; NV2080_CTRL_MC_ARCH_INFO_ARCHITECTURE_*
    NvU32 architecture;
    // Gpu implementation; NV2080_CTRL_MC_ARCH_INFO_IMPLEMENTATION_*
    NvU32 implementation;

    // Host (gpfifo) class; *_CHANNEL_GPFIFO_*, e.g. KEPLER_CHANNEL_GPFIFO_A
    NvU32 host_class;
    // Copy engine (dma) class; *_DMA_COPY_*, e.g. KEPLER_DMA_COPY_A
    NvU32 ce_class;

    // Hardware Abstraction Layer
    uvm_host_hal_t *host_hal;
    uvm_ce_hal_t *ce_hal;
    uvm_arch_hal_t *arch_hal;

    // Whether big page mappings have the physical memory swizzled.
    // On some architectures (Kepler) physical memory mapped with a big page
    // size doesn't follow the common 1-1 mapping where each offset within the
    // big page maps to the same offset within the mapped physical memory.
    // Swizzling makes physically accessing memory mapped with big pages
    // infeasible and has to be worked around by creating virtual identity
    // mappings (virtual identity mappings are a TODO).
    bool big_page_swizzling;

    // Big page size used by the internal UVM VA space
    // Notably it may be different than the big page size used by a user's VA
    // space in general.
    NvU32 internal_big_page_size;

    // RM address space handle used in many of the UVM/RM APIs
    // Represents both an RM device and a GPU VA in RM.
    uvmGpuAddressSpaceHandle rm_address_space;

    // ECC handling
    // In order to trap ECC errors as soon as possible the driver has the hw
    // interrupt register mapped directly. If an ECC interrupt is ever noticed
    // to be pending, then the UVM driver needs to:
    //
    //   1) ask RM to service interrupts, and then
    //   2) inspect the ECC error notifier state.
    //
    // Notably, checking for channel errors is not enough, because ECC errors
    // can be pending, even after a channel has become idle.
    //
    // See more details in uvm_gpu_check_ecc_error().
    struct
    {
        // Does the GPU have ECC enabled?
        bool enabled;

        // Direct mapping of the 32-bit part of the hw interrupt tree that has
        // the ECC bits.
        volatile NvU32 *hw_interrupt_tree_location;

        // Mask to get the ECC interrupt bits from the 32-bits above.
        NvU32 mask;

        // Set to true by RM when a fatal ECC error is encountered (requires
        // asking RM to service pending interrupts to be current).
        NvBool *error_notifier;
    } ecc;

    uvm_gpu_semaphore_pool_t *semaphore_pool;

    uvm_channel_manager_t *channel_manager;

    struct
    {
        // Procfs entry for the GPU directory
        struct proc_dir_entry *dir;

        // Procfs entry for the uuid symlink to the GPU directory
        struct proc_dir_entry *dir_uuid_symlink;

        // Procfs entry for the info file
        struct proc_dir_entry *info_file;
    } procfs;

    uvm_pmm_gpu_t pmm;

    // Protects against changes to the state of a GPU as it transitions from top-half to 
    // bottom-half interrupt handler.
    uvm_mutex_t isr_lock;

    // There is exactly one workqueue per GPU. It is used for the ISR bottom half. Sometimes
    // it is busy, other times it is idle. Note that the Linux kernel has a dedicated kernel
    // thread per CPU, to service workqueues. So N CPUs will be servicing M GPUs, in general.
    struct work_struct workqueue;

    // This is set to true during add_gpu(), if the GPU is Pascal or better. It is set back to
    // false during remove_gpu(). The gpu_table_lock must be held, in order to read or write
    // this variable. This should be treated as a private variable: only those routines, and
    // the ISR code, should read or write this variable.
    bool handling_replayable_faults;
};

typedef enum
{
    UVM_GPU_LINK_INVALID = 0,
    UVM_GPU_LINK_PCIE,
    UVM_GPU_LINK_NVLINK_1,
    UVM_GPU_LINK_MAX
} uvm_gpu_link_type_t;

struct uvm_gpu_peer_struct
{
    // Note: All the peer_caps fields in this global structure can be queried
    // if and only if the corresponding bit from "va_space.enabled_peers" bitmap is set.

    // Peer Id associated with this device w.r.t. to a peer GPU.
    // Note: peerId (A -> B) != peerId (B -> A)
    // peer_id[0] from min(gpu_id_1, gpi_id_2) -> max(gpu_id_1, gpi_id_2)
    // peer_id[1] from max(gpu_id_1, gpi_id_2) -> min(gpu_id_1, gpi_id_2)
    NvU8 peer_ids[2];

    // When this bit is set, peer-to-peer atomics between GPUs are supported
    // natively in hardware instead of being demoted to separate non-atomic
    // read-modify-write accesses.
    NvU8 atomic_supported : 1;

    // The link type between the peer GPUs, currently either PCIE or NVLINK1.
    uvm_gpu_link_type_t link_type;

    NvU64 registered_ref_count;

    // This handle gets populated when enable_peer_access successfully creates
    // an NV50_P2P object. disable_peer_access resets the same on the object deletion.
    NvHandle p2p_handle;
};

// Initialize global gpu state
NV_STATUS uvm_gpu_init(void);

// Deinitialize global state (called from module exit)
void uvm_gpu_exit(void);

// Note that there is a uvm_gpu_get() function defined in uvm_global.h to break
// a circular dep between global and gpu modules.

// Get a gpu by uuid. This returns NULL if the GPU is not present. This is the general purpose
// call that should be used normally.
//
// LOCKING: requires the global lock to be held for read, and also acquires and releases the
// gpu_table_lock.
uvm_gpu_t *uvm_gpu_get_by_uuid(NvProcessorUuid *gpu_uuid);

// Same as uvm_gpu_get_by_uuid, except that this one does not assert that the caller is holding
// the global_lock. This is a narrower purpose function, and is only intended for use by the
// top-half ISR, or other very limited cases.
uvm_gpu_t *uvm_gpu_get_by_uuid_locked(NvProcessorUuid *gpu_uuid);

// Get a registered gpu by uuid. This restricts the search for GPUs, to those that
// have been registered with a va_space. This returns NULL if the GPU is not present, or not
// registered with the va_space.
//
// LOCKING: acquires and releases the gpu_table_lock.
uvm_gpu_t *uvm_gpu_get_registered_by_uuid(uvm_va_space_t *va_space, NvProcessorUuid *gpu_uuid);

// This safely translates a GPU UUID into a GPU ID. If the GPU is not present, or not
// registered, then this returns an error (NV_ERR_INVALID_DEVICE).
static NV_STATUS uvm_registered_uuid_to_id(uvm_va_space_t *va_space,
                                           NvProcessorUuid *gpu_uuid,
                                           uvm_gpu_id_t *id_out)
{
    uvm_gpu_t *gpu = uvm_gpu_get_registered_by_uuid(va_space, gpu_uuid);
    if (!gpu)
        return NV_ERR_INVALID_DEVICE;

    *id_out = gpu->id;
    return NV_OK;
}

// Retain a gpu by uuid
// Returns the retained uvm_gpu_t in gpu_out on success
// LOCKING: requires the global lock to be held for write
NV_STATUS uvm_gpu_retain(NvProcessorUuid *gpu_uuid, uvm_gpu_t **gpu_out);

// Release a gpu
// LOCKING: requires the global lock to be held for write
void uvm_gpu_release(uvm_gpu_t *gpu);

// Decrease the refcount on the GPU object, and actually delete the object if the refcount hits
// zero.
void uvm_gpu_kref_put(uvm_gpu_t *gpu);

// Calculates peer table index using GPU ids.
NvU32 uvm_gpu_peer_table_index(uvm_gpu_id_t gpu_id1, uvm_gpu_id_t gpu_id2);

// Retains existing peer entry. Otherwise, create a new entry.
// LOCKING: requires the global lock to be held for write
NV_STATUS uvm_gpu_retain_peer_access(uvm_gpu_t *gpu_1, uvm_gpu_t *gpu_2);

// Releases a peer entry.
// LOCKING: requires the global lock to be held for write
void uvm_gpu_release_peer_access(uvm_gpu_t *gpu_1, uvm_gpu_t *gpu_2);

// Check for ECC errors
// Notably this check cannot be performed from the top half interrupt handler
NV_STATUS uvm_gpu_check_ecc_error(uvm_gpu_t *gpu);

static bool uvm_gpu_is_gk110_plus(uvm_gpu_t *gpu)
{
    return gpu->architecture >= NV2080_CTRL_MC_ARCH_INFO_ARCHITECTURE_GK110;
}

static bool uvm_processor_uuid_eq(const NvProcessorUuid *uuid1, const NvProcessorUuid *uuid2)
{
    return memcmp(uuid1, uuid2, sizeof(*uuid1)) == 0;
}

// Copies a UUID from source (src) to destination (dst).
static void uvm_processor_uuid_copy(NvProcessorUuid *dst, NvProcessorUuid *src)
{
    memcpy(dst, src, sizeof(*dst));
}

static bool uvm_processor_mask_test(const uvm_processor_mask_t *mask, uvm_processor_id_t id)
{
    UVM_ASSERT_MSG(id < UVM8_MAX_PROCESSORS, "id %u\n", id);

    return test_bit(id, mask->bitmap);
}

// uvm_processor_mask_set and uvm_processor_mask_clear assume to modify the mask
// under a lock, so they use the non-atomic variants __set_bit and __clear_bit.
static void uvm_processor_mask_set(uvm_processor_mask_t *mask, uvm_processor_id_t id)
{
    UVM_ASSERT_MSG(id < UVM8_MAX_PROCESSORS, "id %u\n", id);

    __set_bit(id, mask->bitmap);
}

static void uvm_processor_mask_clear(uvm_processor_mask_t *mask, uvm_processor_id_t id)
{
    UVM_ASSERT_MSG(id < UVM8_MAX_PROCESSORS, "id %u\n", id);

    __clear_bit(id, mask->bitmap);
}

static void uvm_processor_mask_zero(uvm_processor_mask_t *mask)
{
    bitmap_zero(mask->bitmap, UVM8_MAX_PROCESSORS);
}

static bool uvm_processor_mask_empty(uvm_processor_mask_t *mask)
{
    return bitmap_empty(mask->bitmap, UVM8_MAX_PROCESSORS);
}

// Returns the first id that's set in the mask or UVM8_MAX_PROCESSORS if there isn't one.
static uvm_processor_id_t uvm_processor_mask_find_first_id(uvm_processor_mask_t *mask)
{
    return find_first_bit(mask->bitmap, UVM8_MAX_PROCESSORS);
}

// Returns the first id greater or equal to min_id that's set in the mask or
// UVM8_MAX_PROCESSORS if there isn't one.
static uvm_processor_id_t uvm_processor_mask_find_next_id(uvm_processor_mask_t *mask,
                                                          uvm_processor_id_t min_id)
{
    return find_next_bit(mask->bitmap, UVM8_MAX_PROCESSORS, min_id);
}

// Get the number of processors set in the mask
static NvU32 uvm_processor_mask_get_count(uvm_processor_mask_t *mask)
{
    return bitmap_weight(mask->bitmap, UVM8_MAX_PROCESSORS);
}

// Get the number of GPUs set in the mask
static NvU32 uvm_processor_mask_get_gpu_count(uvm_processor_mask_t *mask)
{
    NvU32 gpu_count = uvm_processor_mask_get_count(mask);
    if (uvm_processor_mask_test(mask, UVM_CPU_ID))
        --gpu_count;
    return gpu_count;
}

// Helper to iterate over all processor ids set in a mask
#define for_each_id_in_mask(id, mask)                                       \
    for (id = uvm_processor_mask_find_first_id(mask);                       \
         id != UVM8_MAX_PROCESSORS;                                         \
         id = uvm_processor_mask_find_next_id(mask, id + 1))

// Helper to iterate over all GPU ids set in a mask
#define for_each_gpu_id_in_mask(gpu_id, mask)                               \
    for (gpu_id = uvm_processor_mask_find_next_id(mask, UVM_CPU_ID + 1);    \
         gpu_id != UVM8_MAX_PROCESSORS;                                     \
         gpu_id = uvm_processor_mask_find_next_id(mask, gpu_id + 1))

#endif // __UVM8_GPU_H__
