/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "nv_uvm_interface.h"
#include "uvm8_global.h"
#include "uvm8_gpu_page_fault.h"
#include "uvm_common.h"

// Convenient counters for non-intrusive monitoring:
static atomic64_t g_uvm8_interrupt_count_total = ATOMIC64_INIT(0);
static atomic64_t g_uvm8_interrupt_count_bottom_half = ATOMIC64_INIT(0);

static NV_STATUS uvm_service_faultbuffer(uvm_gpu_t *gpu);
static void uvm8_isr_bottom_half(struct work_struct *work);

// This is called from RM's top-half ISR (see: the nvidia_isr() function), and UVM is given a
// chance to handle the interrupt, before most of the RM processing. UVM communicates what it
// did, back to RM, via the return code:
//
//     NV_OK:
//         UVM handled the interrupt, and RM does not need to schedule its own bottom half.
//
//     NV_WARN_MORE_PROCESSING_REQUIRED:
//         UVM did not schedule a bottom half, because it was unable to get the locks it
//         needed, but there is still UVM work to be done. RM will return "not handled" to the
//         Linux kernel, *unless* RM handled other faults in its top half. In that case, the
//         fact that UVM did not handle its interrupt is lost. However, life and interrupt
//         processing continues anyway: the GPU will soon raise another interrupt, because
//         that's what it does when there are replayable page faults remaining (GET != PUT in
//         the fault buffer).
//
//     NV_ERR_NO_INTR_PENDING:
//         UVM did not find any work to do. Currently this is handled in RM in exactly the same
//         way as NV_WARN_MORE_PROCESSING_REQUIRED is handled. However, the extra precision is
//         available for the future. RM's interrupt handling tends to evolve as new chips and
//         new interrupts get created.

NV_STATUS uvm8_isr_top_half(NvProcessorUuid *gpu_uuid)
{
    uvm_gpu_t *gpu;
    NV_STATUS status = NV_ERR_NO_INTR_PENDING;

    if (!in_interrupt()) {
        // Early-out if not in interrupt context. This happens with
        // CONFIG_DEBUG_SHIRQ enabled where the interrupt handler is called as
        // part of its removal to make sure it's prepared for being called even
        // when it's being freed.
        // This breaks the assumption that the UVM driver is called in atomic
        // context only in the interrupt context, which
        // uvm_thread_context_retain() relies on.
        return NV_OK;
    }

    atomic64_inc(&g_uvm8_interrupt_count_total);

    if (!gpu_uuid) {
        // This can happen early in the main GPU driver initialization, because
        // that involves testing interrupts before the GPU is fully set up.
        return status;
    }

    uvm_spin_lock_irqsave(&g_uvm_global.gpu_table_lock);

    gpu = uvm_gpu_get_by_uuid_locked(gpu_uuid);

    if (gpu == NULL || !gpu->handling_replayable_faults) {
        uvm_spin_unlock_irqrestore(&g_uvm_global.gpu_table_lock);
        return status;
    }

    kref_get(&gpu->gpu_kref);

    // TODO: add support to lockdep, for leaving this lock acquired (the bottom half eventually
    // releases it).
    if (mutex_trylock(&gpu->isr_lock.m) == 0) {
        status = NV_WARN_MORE_PROCESSING_REQUIRED;
        goto done_no_bottom_half;
    }

    if (!uvm_replayable_faults_pending(gpu))
        goto done_no_bottom_half;

    uvm_disable_pagefault_interrupts(gpu);

    // Schedule a bottom half, but do *not* release the GPU ISR lock. The bottom half releases
    // the GPU ISR lock as part of its cleanup.
    NV_WORKQUEUE_INIT_AND_SCHEDULE(&gpu->workqueue, uvm8_isr_bottom_half);

    uvm_spin_unlock_irqrestore(&g_uvm_global.gpu_table_lock);

    // Keep the isr_lock, and the gpu_kref count, and run the bottom half:
    return NV_OK;

done_no_bottom_half:
    uvm_spin_unlock_irqrestore(&g_uvm_global.gpu_table_lock);
    uvm_gpu_kref_put(gpu);

    return status;
}

static void uvm8_isr_bottom_half(struct work_struct *work)
{
    uvm_gpu_t *gpu = container_of(work, uvm_gpu_t, workqueue);

    atomic64_inc(&g_uvm8_interrupt_count_bottom_half);

    // TODO (here, or within uvm_service_faultbuffer): acquire and release the va_space lock.
    uvm_service_faultbuffer(gpu);

    uvm_spin_lock_irqsave(&g_uvm_global.gpu_table_lock);

    // If this is false, that means that remove_gpu() has already removed this GPU from the GPU
    // table, and it is on its way down.
    if (gpu->handling_replayable_faults)
        uvm_enable_pagefault_interrupts(gpu);

    // Raw unlock call, to correspond to the raw lock call in the top half:
    mutex_unlock(&gpu->isr_lock.m);

    uvm_spin_unlock_irqrestore(&g_uvm_global.gpu_table_lock);

    uvm_gpu_kref_put(gpu);
}

bool uvm_replayable_faults_pending(uvm_gpu_t *gpu)
{
    // TODO: implement. Check fast path (faultbuffer, in sysmem) first, then slow path
    // (GPU register via BAR0 over PCIe).
    //
    // For now, provide more testing coverage by returning true. This will cause a bottom
    // half to run:
    return true;
}

static NV_STATUS uvm_service_faultbuffer(uvm_gpu_t *gpu)
{
    // TODO: implement
    return NV_OK;
}

void uvm_enable_pagefault_interrupts(uvm_gpu_t *gpu)
{
    // TODO: implement
}

void uvm_disable_pagefault_interrupts(uvm_gpu_t *gpu)
{
    // TODO: implement
}


