/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm8_hal.h"
#include "uvm8_push.h"
#include "cla06f.h"
#include "cla16f.h"

void uvm_hal_kepler_host_wait_for_idle_a06f(uvm_push_t *push)
{
    NV_PUSH_1U(A06F, SET_REFERENCE, 0);
}

void uvm_hal_kepler_host_wait_for_idle_a16f(uvm_push_t *push)
{
    NV_PUSH_1U(A16F, WFI, 0);
}

void uvm_hal_kepler_host_membar_sys(uvm_push_t *push)
{
    NV_PUSH_1U(A06F, MEM_OP_B,
       HWCONST(A06F, MEM_OP_B, OPERATION, SYSMEMBAR_FLUSH));
}

void uvm_hal_kepler_host_noop(uvm_push_t *push, NvU32 size)
{
    UVM_ASSERT_MSG(size % 4 == 0, "size %u\n", size);

    if (size == 0)
        return;

    // size is in bytes so divide by the method size (4 bytes)
    size /= 4;

    while (size > 0) {
        // noop_this_time includes the NOP method itself and hence can be
        // up to COUNT_MAX + 1.
        NvU32 noop_this_time = min(UVM_METHOD_COUNT_MAX + 1, size);

        // -1 for the NOP method itself.
        NV_PUSH_NU_NONINC(A06F, NOP, noop_this_time - 1);

        size -= noop_this_time;
    }
}

void uvm_hal_kepler_host_semaphore_release(uvm_push_t *push, uvm_gpu_semaphore_t *semaphore, NvU32 payload)
{
    uvm_gpu_t *gpu = uvm_push_get_gpu(push);
    NvU64 semaphore_va = uvm_gpu_semaphore_get_gpu_va(semaphore, gpu);
    NV_PUSH_4U(A06F, SEMAPHOREA, NvOffset_HI32(semaphore_va),
                     SEMAPHOREB, NvOffset_LO32(semaphore_va),
                     SEMAPHOREC, payload,
                     SEMAPHORED, HWCONST(A06F, SEMAPHORED, OPERATION, RELEASE) |
                                 HWCONST(A06F, SEMAPHORED, RELEASE_SIZE, 4BYTE)|
                                 HWCONST(A06F, SEMAPHORED, RELEASE_WFI, DIS));
}

void uvm_hal_kepler_host_semaphore_acquire(uvm_push_t *push, uvm_gpu_semaphore_t *semaphore, NvU32 payload)
{
    uvm_gpu_t *gpu = uvm_push_get_gpu(push);
    NvU64 semaphore_va = uvm_gpu_semaphore_get_gpu_va(semaphore, gpu);
    NV_PUSH_4U(A06F, SEMAPHOREA, NvOffset_HI32(semaphore_va),
                     SEMAPHOREB, NvOffset_LO32(semaphore_va),
                     SEMAPHOREC, payload,
                     SEMAPHORED, HWCONST(A06F, SEMAPHORED, ACQUIRE_SWITCH, ENABLED) |
                                 HWCONST(A06F, SEMAPHORED, OPERATION, ACQ_GEQ));

}

void uvm_hal_kepler_host_set_gpfifo_entry(NvU64 *fifo_entry, NvU64 pushbuffer_va, NvU32 pushbuffer_length)
{
    NvU64 fifo_entry_value;

    UVM_ASSERT_MSG(pushbuffer_va % 4 == 0, "pushbuffer va unaligned: %llu\n", pushbuffer_va);
    UVM_ASSERT_MSG(pushbuffer_length % 4 == 0, "pushbuffer length unaligned: %u\n", pushbuffer_length);

    fifo_entry_value =          HWVALUE(A06F, GP_ENTRY0, GET, NvU64_LO32(pushbuffer_va) >> 2);
    fifo_entry_value |= (NvU64)(HWVALUE(A06F, GP_ENTRY1, GET_HI, NvU64_HI32(pushbuffer_va)) |
                                HWVALUE(A06F, GP_ENTRY1, LENGTH, pushbuffer_length >> 2) |
                                HWCONST(A06F, GP_ENTRY1, PRIV,   KERNEL)) << 32;

    *fifo_entry = fifo_entry_value;
}
