/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvmtypes.h"
#include "uvm8_forward_decl.h"
#include "uvm8_gpu.h"
#include "uvm8_mmu.h"
#include "uvm8_hal.h"
#include "uvm8_kvmalloc.h"
#include "uvm8_push.h"

// The page tree has 5 levels on pascal, and the root is never freed by a normal 'put' operation
// which leaves a maximum of 4 levels
#define MAX_OPERATION_DEPTH 4

static NV_STATUS phys_mem_allocate(uvm_gpu_t *gpu, NvLength size, uvm_aperture_t location, uvm_mmu_page_table_alloc_t *out)
{
    // TODO: support UVM_APERTURE_VID
    UVM_ASSERT(location == UVM_APERTURE_SYS);

    memset(out, 0, sizeof(*out));
    out->handle.page = alloc_pages(NV_UVM_GFP_FLAGS | __GFP_ZERO, get_order(size));
    if (out->handle.page == NULL)
        return NV_ERR_NO_MEMORY;

    out->size = size;
    out->aperture = location;

    // TODO: this address needs to be the physical address from the GPU's perspective
    // which will be different when there's an IOMMU
    out->addr = page_to_phys(out->handle.page);
    return NV_OK;
}

static void phys_mem_deallocate(uvm_gpu_t *gpu, uvm_mmu_page_table_alloc_t *ptr)
{
    UVM_ASSERT(ptr->aperture == UVM_APERTURE_SYS);
    __free_pages(ptr->handle.page, get_order(ptr->size));
    memset(ptr, 0, sizeof(*ptr));
}

static void phys_mem_init(uvm_page_tree_t *tree, NvU32 page_size, uvm_page_directory_t *dir, uvm_push_t *push)
{
    NvU64 clear_bits[2];
    uvm_mmu_mode_hal_t *hal = tree->hal;

    if (dir->depth == tree->hal->page_table_depth(page_size)) {
        *clear_bits = tree->hal->empty_pte(page_size);
    }
    else {
        // passing in NULL for the phys_allocs will result in a the correct PDE bit pattern
        uvm_mmu_page_table_alloc_t *phys_allocs[2] = {NULL, NULL};
        hal->make_pde(clear_bits, phys_allocs, dir->depth);

        // Make sure that using only clear_bits[0] will work
        UVM_ASSERT(hal->entry_size(dir->depth) == sizeof(clear_bits[0]) || clear_bits[0] == clear_bits[1]);
    }

    // initialize the memory to a reasonable value
    tree->gpu->ce_hal->memset_8(push,
                                uvm_gpu_address_physical(dir->phys_alloc.aperture, dir->phys_alloc.addr),
                                *clear_bits,
                                dir->phys_alloc.size);
}

static uvm_page_directory_t *allocate_directory(uvm_page_tree_t *tree, NvU32 page_size, uvm_page_directory_t *parent)
{
    NV_STATUS status;
    uvm_mmu_mode_hal_t *hal = tree->hal;
    NvU32 depth = (parent == NULL) ? 0 : parent->depth + 1;
    NvU32 entry_count;
    NvLength phys_alloc_size = hal->allocation_size(depth, page_size);
    uvm_page_directory_t *dir;

    // The page tree doesn't cache PTEs so space is not allocated for entries that are always PTEs.
    // 2M PTEs may later become PDEs so pass UVM_PAGE_SIZE_AGNOSTIC, not page_size.
    if (depth == hal->page_table_depth(UVM_PAGE_SIZE_AGNOSTIC))
        entry_count = 0;
    else
        entry_count = hal->entries_per_index(depth) << hal->index_bits(depth, page_size);

    // TODO: use a slab allocator for this?
    dir = uvm_kvmalloc_zero(sizeof(uvm_page_directory_t) + sizeof(dir->entries[0]) * entry_count);
    if (dir == NULL)
        return NULL;

    status = phys_mem_allocate(tree->gpu, phys_alloc_size, UVM_APERTURE_SYS, &dir->phys_alloc);
    if (status != NV_OK) {
        uvm_kvfree(dir);
        return NULL;
    }

    dir->host_parent = parent;
    dir->depth = depth;

    return dir;
}

static inline NvU32 entry_index_from_vaddr(NvU64 vaddr, NvU32 addr_bit_shift, NvU32 bits)
{
    NvU64 mask = ((NvU64)1 << bits) - 1;
    return (NvU32)((vaddr >> addr_bit_shift) & mask);
}

static inline NvU32 index_to_entry(uvm_mmu_mode_hal_t *hal, NvU32 entry_index, NvU32 depth, NvU32 page_size)
{
    return hal->entries_per_index(depth) * entry_index + hal->entry_offset(depth, page_size);
}

static void pde_write(uvm_page_tree_t *tree, uvm_page_directory_t *dir, NvU32 entry_index, uvm_push_t *push)
{
    NvU32 i;
    NvU64 entry_bits[2];
    uvm_mmu_page_table_alloc_t *phys_allocs[2];
    uvm_mmu_mode_hal_t *hal = tree->hal;
    NvU64 dev_entry = dir->phys_alloc.addr + hal->entry_size(dir->depth) * entry_index;
    NvU32 entries_per_index = hal->entries_per_index(dir->depth);
    NvU32 entry_size = hal->entry_size(dir->depth);

    UVM_ASSERT(sizeof(entry_bits) >= entry_size);

    // extract physical allocs from non-null entries
    for (i = 0; i < entries_per_index; i++) {
        uvm_page_directory_t *entry = dir->entries[entries_per_index * entry_index + i];
        if (entry == NULL)
            phys_allocs[i] = NULL;
        else
            phys_allocs[i] = &entry->phys_alloc;
    }

    // make_pde always writes the whole PDE, even it is a dual entry.
    hal->make_pde(entry_bits, phys_allocs, dir->depth);

    // each entry could be either 1 or 2 64-bit words.
    for (i = 0; i < entry_size / sizeof(entry_bits[0]); i++) {
        tree->gpu->ce_hal->memset_8(push,
                uvm_gpu_address_physical(dir->phys_alloc.aperture, dev_entry + sizeof(entry_bits[0]) * i),
                entry_bits[i],
                sizeof(entry_bits[0]));
    }
}

static void host_pde_clear(uvm_page_tree_t *tree, uvm_page_directory_t *dir, NvU32 entry_index, NvU32 page_size)
{
    UVM_ASSERT(dir->ref_count > 0);

    dir->entries[index_to_entry(tree->hal, entry_index, dir->depth, page_size)] = NULL;
    dir->ref_count--;
}

static void pde_clear(uvm_page_tree_t *tree, uvm_page_directory_t *dir, NvU32 entry_index, NvU32 page_size, void *push)
{
    host_pde_clear(tree, dir, entry_index, page_size);
    pde_write(tree, dir, entry_index, push);
}

static uvm_chunk_sizes_mask_t allocation_sizes_for_big_page_size(uvm_gpu_t *gpu, NvU32 big_page_size)
{
    uvm_chunk_sizes_mask_t alloc_sizes = 0;
    uvm_mmu_mode_hal_t *hal = gpu->arch_hal->mmu_mode_hal(big_page_size);

    if (hal != NULL) {
        unsigned long page_size_log2;
        unsigned long page_sizes = hal->page_sizes();
        BUILD_BUG_ON(sizeof(hal->page_sizes()) > sizeof(page_sizes));

        for_each_set_bit(page_size_log2, &page_sizes, BITS_PER_LONG) {
            NvU32 i;
            uvm_chunk_size_t page_size = (uvm_chunk_size_t)1 << page_size_log2;
            for (i = 0; i <= hal->page_table_depth(page_size); i++)
                alloc_sizes |= hal->allocation_size(i, page_size);
        }
    }

    return alloc_sizes;
}

static NvU32 page_sizes_for_big_page_size(uvm_gpu_t *gpu, NvU32 big_page_size)
{
    uvm_mmu_mode_hal_t *hal = gpu->arch_hal->mmu_mode_hal(big_page_size);

    if (hal != NULL)
        return hal->page_sizes();
    else
        return 0;
}

NV_STATUS uvm_page_tree_init(uvm_gpu_t *gpu, NvU32 big_page_size, uvm_page_tree_t *tree, uvm_push_t *push)
{
    BUILD_BUG_ON(sizeof(uvm_page_directory_t) != offsetof(uvm_page_directory_t, entries));

    memset(tree, 0, sizeof(*tree));

    tree->hal = gpu->arch_hal->mmu_mode_hal(big_page_size);
    UVM_ASSERT(tree->hal != NULL);
    tree->gpu = gpu;
    tree->big_page_size = big_page_size;

    tree->root = allocate_directory(tree, UVM_PAGE_SIZE_AGNOSTIC, NULL);
    if (tree->root == NULL)
        return NV_ERR_NO_MEMORY;

    phys_mem_init(tree, UVM_PAGE_SIZE_AGNOSTIC, tree->root, push);
    return NV_OK;
}

void uvm_page_tree_deinit(uvm_page_tree_t *tree)
{
    UVM_ASSERT(tree->root->ref_count == 0);

    phys_mem_deallocate(tree->gpu, &tree->root->phys_alloc);
    uvm_kvfree(tree->root);
}

void uvm_page_tree_put_ptes(uvm_page_tree_t *tree, uvm_page_table_range_t *range, uvm_tracker_t *deps)
{
    NvU32 free_count = 0;
    NvU32 i;
    uvm_page_directory_t *free_queue[MAX_OPERATION_DEPTH];
    uvm_page_directory_t *dir = range->table;
    uvm_push_t push;
    NvBool test_mode = tree->gpu->channel_manager == NULL;

    // begin a push which will be submitted before the memory gets freed
    if (!test_mode) {
        NV_STATUS status;
        status = uvm_push_begin_acquire(tree->gpu->channel_manager,
                                        UVM_CHANNEL_TYPE_MEMOPS,
                                        deps,
                                        &push,
                                        "put_ptes: range start: %u, length: %u",
                                        range->start_index,
                                        range->entry_count);

        if (status != NV_OK)
            return;
    }

    // release the range
    UVM_ASSERT(dir->ref_count >= range->entry_count);
    dir->ref_count -= range->entry_count;

    // traverse until we hit an in-use page, or the root
    while (dir->host_parent != NULL && dir->ref_count == 0) {
        uvm_page_directory_t *parent = dir->host_parent;

        pde_clear(tree, dir->host_parent, dir->index_in_parent, range->page_size, &push);

        // add this dir to the queue of directories that should be freed once the
        // tracker value of the associated PDE writes is known
        UVM_ASSERT(free_count < MAX_OPERATION_DEPTH);
        free_queue[free_count++] = dir;

        dir = parent;
    }

    // TODO: invalidate now that all the PDE clears have been pushed
    // wait for the pde_clear's to go through so it's safe to free everything
    if (!test_mode)
        uvm_push_end_and_wait(&push);

    // now that we've traversed all the way up the tree, free everything
    for (i = 0; i < free_count; i++) {
        phys_mem_deallocate(tree->gpu, &free_queue[i]->phys_alloc);
        uvm_kvfree(free_queue[i]);
    }
}

NV_STATUS uvm_page_tree_get_ptes(uvm_page_tree_t *tree, NvU32 page_size, NvU64 start, NvLength size, uvm_push_t *push, uvm_page_table_range_t *range)
{
    NV_STATUS status = NV_OK;
    NvU32 i;
    uvm_mmu_mode_hal_t *hal = tree->hal;
    // bit index just beyond the most significant bit used to index the current entry
    NvU32 addr_bit_shift;
    // track level upon which the invalidate occured
    NvS32 invalidate_level = -1;
    uvm_page_directory_t *dir = tree->root;

    // array of entries that have been allocated this time
    uvm_page_directory_t *alloc_stack[MAX_OPERATION_DEPTH];
    NvU32 alloc_count = 0;

    // ensure that the caller has specified a valid page size
    UVM_ASSERT((page_size & hal->page_sizes()) != 0);

    // This algorithm will work with unaligned ranges, but the caller's intent is unclear
    UVM_ASSERT(start % page_size == 0 && size % page_size == 0);

    addr_bit_shift = hal->num_va_bits();

    while (true) {
        // index of the entry, for the first byte of the range, within its containing directory
        NvU32 start_index;
        // index of the entry, for the last byte of the range, within its containing directory
        NvU32 end_index;
        // pointer to PDE/PTE
        uvm_page_directory_t **entry;
        NvU32 index_bits = hal->index_bits(dir->depth, page_size);

        addr_bit_shift -= index_bits;
        start_index = entry_index_from_vaddr(start, addr_bit_shift, index_bits);
        end_index = entry_index_from_vaddr(start + size - 1, addr_bit_shift, index_bits);

        UVM_ASSERT(start_index <= end_index && end_index < (1 << index_bits));

        entry = dir->entries + index_to_entry(hal, start_index, dir->depth, page_size);

        if (dir->depth == hal->page_table_depth(page_size)) {
            range->table = dir;
            range->start_index = start_index;
            range->entry_count = 1 + end_index - start_index;
            range->page_size = page_size;
            dir->ref_count += range->entry_count;
            break;
        }
        else {
            UVM_ASSERT(start_index == end_index);

            if (*entry == NULL) {
                *entry = allocate_directory(tree, page_size, dir);
                if (*entry == NULL) {
                    status = NV_ERR_NO_MEMORY;
                    goto error;
                }

                (*entry)->index_in_parent = start_index;
                dir->ref_count++;

                alloc_stack[alloc_count++] = *entry;

                if (invalidate_level == -1)
                    invalidate_level = dir->depth;
            }
        }
        dir = *entry;
    }

    // only do GPU work once all the allocations have succeeded
    // first, zero-out the new allocations
    for (i = 0; i < alloc_count; i++)
        phys_mem_init(tree, page_size, alloc_stack[i], push);

    // write entries bottom up, so that they are valid once they're inserted into the tree
    i = alloc_count;
    while (i > 0) {
        uvm_page_directory_t *dir = alloc_stack[--i];
        pde_write(tree, dir->host_parent, dir->index_in_parent, push);
    }

    if (invalidate_level >= 0)
        /* TODO: invalidate */;
    return NV_OK;

error:
    // put the tree back to its previous state on error
    while (alloc_count > 0) {
        uvm_page_directory_t *dir = alloc_stack[--alloc_count];
        host_pde_clear(tree, dir->host_parent, dir->index_in_parent, page_size);
        phys_mem_deallocate(tree->gpu, &dir->phys_alloc);
        uvm_kvfree(dir);
    }
    return status;
}

uvm_mmu_page_table_alloc_t *uvm_page_tree_pdb(uvm_page_tree_t *tree)
{
    return &tree->root->phys_alloc;
}

uvm_chunk_sizes_mask_t uvm_mmu_kernel_chunk_sizes(uvm_gpu_t *gpu)
{
    return allocation_sizes_for_big_page_size(gpu, UVM_PAGE_SIZE_64K) |
           allocation_sizes_for_big_page_size(gpu, UVM_PAGE_SIZE_128K);
}

uvm_chunk_sizes_mask_t uvm_mmu_user_chunk_sizes(uvm_gpu_t *gpu)
{
    return page_sizes_for_big_page_size(gpu, UVM_PAGE_SIZE_64K) |
           page_sizes_for_big_page_size(gpu, UVM_PAGE_SIZE_128K);
}
