/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm8_test.h"
#include "uvm8_test_ioctl.h"
#include "uvm8_gpu.h"
#include "uvm8_hal.h"
#include "uvm8_mmu.h"
// KEPLER_*
#include "cla0b5.h"
#include "cla06f.h"
// PASCAL_*
#include "clc0b5.h"
#include "clc06f.h"
// ARCHITECTURE_*
#include "ctrl2080mc.h"

#define BIG_PAGE_SIZE_PASCAL (1 << 16)

static void fake_ce_memset_8(uvm_push_t *push, uvm_gpu_address_t dst, NvU64 value, size_t size)
{
    size_t i;

    UVM_ASSERT(dst.aperture == UVM_APERTURE_SYS);

    for (i = 0; i < size; i += 8)
        *(NvU64 *)phys_to_virt(dst.address + i) = value;
}

static NV_STATUS allocate_root(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, BIG_PAGE_SIZE_PASCAL, &tree, NULL), NV_OK);
    uvm_page_tree_deinit(&tree);
    return NV_OK;
}

static NV_STATUS alloc_64k_memory(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range;

    NvLength size = 64 * 1024;
    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, BIG_PAGE_SIZE_PASCAL, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_64K, 0, size, NULL, &range), NV_OK);
    TEST_CHECK_RET(range.entry_count == 1);
    TEST_CHECK_RET(range.table->depth == 4);
    TEST_CHECK_RET(range.start_index == 0);
    TEST_CHECK_RET(range.page_size == UVM_PAGE_SIZE_64K);
    TEST_CHECK_RET(tree.root->ref_count == 1);
    TEST_CHECK_RET(tree.root->entries[0]->ref_count == 1);
    TEST_CHECK_RET(tree.root->entries[0]->entries[0]->ref_count == 1);
    TEST_CHECK_RET(tree.root->entries[0]->entries[0]->entries[0]->ref_count == 1);
    TEST_CHECK_RET(tree.root->entries[0]->entries[0]->entries[0]->entries[0]->ref_count == 1);
    TEST_CHECK_RET(range.table == tree.root->entries[0]->entries[0]->entries[0]->entries[0]);
    uvm_page_tree_put_ptes(&tree, &range, NULL);
    UVM_ASSERT(tree.root->ref_count == 0);
    uvm_page_tree_deinit(&tree);
    return NV_OK;
}

static NV_STATUS alloc_adjacent_64k_memory(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range1;
    uvm_page_table_range_t range2;

    NvLength size = 64 * 1024;
    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, BIG_PAGE_SIZE_PASCAL, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_64K, size, size, NULL, &range1), NV_OK);
    TEST_CHECK_RET(range1.entry_count == 1);

    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_64K, 0, size, NULL, &range2), NV_OK);
    TEST_CHECK_RET(range2.entry_count == 1);
    TEST_CHECK_RET(range1.table == range2.table);
    TEST_CHECK_RET(range1.table == tree.root->entries[0]->entries[0]->entries[0]->entries[0]);
    TEST_CHECK_RET(range1.start_index == 1);
    TEST_CHECK_RET(range2.start_index == 0);

    uvm_page_tree_put_ptes(&tree, &range1, NULL);
    uvm_page_tree_put_ptes(&tree, &range2, NULL);
    uvm_page_tree_deinit(&tree);
    return NV_OK;
}

static NV_STATUS alloc_adjacent_pde_64k_memory(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range;
    uvm_page_table_range_t next_range;
    NvLength size = 64 * 1024;

    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, BIG_PAGE_SIZE_PASCAL, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_64K, 0, size, NULL, &range), NV_OK);
    TEST_CHECK_RET(range.entry_count == 1);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_64K, 2 * 1024 * 1024, size, NULL, &next_range), NV_OK);
    TEST_CHECK_RET(range.table == tree.root->entries[0]->entries[0]->entries[0]->entries[0]);
    TEST_CHECK_RET(next_range.table == tree.root->entries[0]->entries[0]->entries[0]->entries[2]);
    uvm_page_tree_put_ptes(&tree, &range, NULL);
    uvm_page_tree_put_ptes(&tree, &next_range, NULL);
    uvm_page_tree_deinit(&tree);
    return NV_OK;
}


static NV_STATUS alloc_nearby_pde_64k_memory(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range;
    uvm_page_table_range_t next_range;
    NvLength size = 64 * 1024;
    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, BIG_PAGE_SIZE_PASCAL, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_64K, 6 * 1024 * 1024, size, NULL, &range), NV_OK);
    TEST_CHECK_RET(range.entry_count == 1);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_64K, 2 * 1024 * 1024, size, NULL, &next_range), NV_OK);
    TEST_CHECK_RET(range.table == tree.root->entries[0]->entries[0]->entries[0]->entries[6]);
    TEST_CHECK_RET(next_range.table == tree.root->entries[0]->entries[0]->entries[0]->entries[2]);
    uvm_page_tree_put_ptes(&tree, &range, NULL);
    uvm_page_tree_put_ptes(&tree, &next_range, NULL);
    uvm_page_tree_deinit(&tree);
    return NV_OK;
}

static NV_STATUS allocate_then_free_all_16_64k(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range[16];

    NvLength size = 64 * 1024;
    NvLength stride = 32 * size;
    NvLength start = stride * 256;
    int i;

    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, BIG_PAGE_SIZE_PASCAL, &tree, NULL), NV_OK);

    for (i = 0; i < 16; i++)
        MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_64K, start + i * stride, size, NULL, range + i), NV_OK);

    TEST_CHECK_RET(tree.root->entries[0]->entries[0]->entries[1]->ref_count == 16);

    for (i = 0; i < 16; i++)
        uvm_page_tree_put_ptes(&tree, range + i, NULL);

    UVM_ASSERT(tree.root->ref_count == 0);
    uvm_page_tree_deinit(&tree);
    return NV_OK;
}

static NV_STATUS allocate_then_free_8_8_64k(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range[16];

    NvLength size = 64 * 1024;
    NvLength stride = 32 * size;
    NvLength start = stride * 248 + 256LL * 1024 * 1024 * 1024 + (1LL << 47);
    int i;

    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, BIG_PAGE_SIZE_PASCAL, &tree, NULL), NV_OK);

    for (i = 0; i < 16; i++)
        MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_64K, start + i * stride , size, NULL, range + i), NV_OK);

    TEST_CHECK_RET(tree.root->entries[1]->entries[1]->entries[0]->ref_count == 8);
    TEST_CHECK_RET(tree.root->entries[1]->entries[1]->entries[1]->ref_count == 8);

    for (i = 0; i < 16; i++)
        uvm_page_tree_put_ptes(&tree, range + i, NULL);

    UVM_ASSERT(tree.root->ref_count == 0);
    uvm_page_tree_deinit(&tree);
    return NV_OK;
}

static NV_STATUS get_single_page_2m(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range;

    // use a start address not at the beginning of a PDE3 entry's range
    NvU64 start = 34983UL * (1 << 21);
    NvLength size = 1 << 21;

    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, BIG_PAGE_SIZE_PASCAL, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_2M, start, size, NULL, &range), NV_OK);

    TEST_CHECK_RET(range.entry_count == 1);
    TEST_CHECK_RET(range.table->depth == 3);
    TEST_CHECK_RET(range.page_size == UVM_PAGE_SIZE_2M);

    uvm_page_tree_put_ptes(&tree, &range, NULL);
    TEST_CHECK_RET(tree.root->ref_count == 0);
    uvm_page_tree_deinit(&tree);
    return NV_OK;
}


static NV_STATUS get_entire_table_4k(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range;

    NvU64 start = 1UL << 47;

    NvLength size = 1 << 21;

    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, BIG_PAGE_SIZE_PASCAL, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_4K, start, size, NULL, &range), NV_OK);

    TEST_CHECK_RET(range.table == tree.root->entries[1]->entries[0]->entries[0]->entries[1]);
    TEST_CHECK_RET(range.entry_count == 512);
    TEST_CHECK_RET(range.table->depth == 4);
    TEST_CHECK_RET(range.page_size == UVM_PAGE_SIZE_4K);
    TEST_CHECK_RET(tree.root->ref_count == 1);

    uvm_page_tree_put_ptes(&tree, &range, NULL);
    uvm_page_tree_deinit(&tree);
    return NV_OK;
}

static NV_STATUS split_4k_from_2m(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range_2m;
    uvm_page_table_range_t range_adj;
    uvm_page_table_range_t range_4k;
    uvm_page_table_range_t range_64k;

    NvU64 start = 1UL << 48;
    NvLength size = 1 << 21;

    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, BIG_PAGE_SIZE_PASCAL, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_2M, start, size, NULL, &range_2m), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_2M, start + size, size, NULL, &range_adj), NV_OK);

    TEST_CHECK_RET(range_2m.entry_count == 1);
    TEST_CHECK_RET(range_2m.table->depth == 3);
    TEST_CHECK_RET(range_adj.entry_count == 1);
    TEST_CHECK_RET(range_adj.table->depth == 3);

    // Need to release the 2 MB page so that the reference count is right.
    uvm_page_tree_put_ptes(&tree, &range_2m, NULL);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_4K, start, 64 * 1024, NULL, &range_4k), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_64K, start + 64 * 1024, size - 64 * 1024, NULL, &range_64k), NV_OK);

    TEST_CHECK_RET(range_4k.entry_count == 16);
    TEST_CHECK_RET(range_4k.table->depth == 4);
    TEST_CHECK_RET(range_4k.table == tree.root->entries[2]->entries[0]->entries[0]->entries[1]);
    TEST_CHECK_RET(range_4k.start_index == 0);

    TEST_CHECK_RET(range_64k.entry_count == 31);
    TEST_CHECK_RET(range_64k.table == tree.root->entries[2]->entries[0]->entries[0]->entries[0]);
    TEST_CHECK_RET(range_64k.start_index == 1);

    // Free everything
    uvm_page_tree_put_ptes(&tree, &range_adj, NULL);
    uvm_page_tree_put_ptes(&tree, &range_4k, NULL);
    uvm_page_tree_put_ptes(&tree, &range_64k, NULL);

    uvm_page_tree_deinit(&tree);
    return NV_OK;
}

static NV_STATUS get_512mb_range(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range;

    NvU64 start = 512 * (1 << 20);
    NvU64 size = start;

    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, BIG_PAGE_SIZE_PASCAL, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_2M, start, size, NULL, &range), NV_OK);
    TEST_CHECK_RET(range.entry_count == 256);
    TEST_CHECK_RET(range.table->depth == 3);
    TEST_CHECK_RET(range.start_index == 0);
    uvm_page_tree_put_ptes(&tree, &range, NULL);
    uvm_page_tree_deinit(&tree);
    return NV_OK;
}

static NV_STATUS get_two_free_apart(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range1;
    uvm_page_table_range_t range2;

    NvLength size = 1024 * 1024;
    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, BIG_PAGE_SIZE_PASCAL, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_4K, size, size, NULL, &range1), NV_OK);
    TEST_CHECK_RET(range1.entry_count == 256);
    TEST_CHECK_RET(range1.table->ref_count == 256);

    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_4K, 0, size, NULL, &range2), NV_OK);
    TEST_CHECK_RET(range2.entry_count == 256);
    TEST_CHECK_RET(range2.table->ref_count == 512);
    TEST_CHECK_RET(range1.table == range2.table);
    // 4k page is second entry in a dual PDE
    TEST_CHECK_RET(range1.table == tree.root->entries[0]->entries[0]->entries[0]->entries[1]);
    TEST_CHECK_RET(range1.start_index == 256);
    TEST_CHECK_RET(range2.start_index == 0);

    uvm_page_tree_put_ptes(&tree, &range1, NULL);
    TEST_CHECK_RET(range2.table->ref_count == 256);
    TEST_CHECK_RET(range2.table == tree.root->entries[0]->entries[0]->entries[0]->entries[1]);
    uvm_page_tree_put_ptes(&tree, &range2, NULL);
    uvm_page_tree_deinit(&tree);
    return NV_OK;

}

static NV_STATUS get_overlapping_dual_pdes(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range4k;
    uvm_page_table_range_t range64k;

    NvLength size = 1024 * 1024;
    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, BIG_PAGE_SIZE_PASCAL, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_4K, size, size, NULL, &range4k), NV_OK);
    TEST_CHECK_RET(range4k.entry_count == 256);
    TEST_CHECK_RET(range4k.table->ref_count == 256);

    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_64K, size, size, NULL, &range64k), NV_OK);
    TEST_CHECK_RET(range64k.entry_count == 16);
    TEST_CHECK_RET(range64k.table->ref_count == 16);
    // 4k page is second entry in a dual PDE
    TEST_CHECK_RET(range64k.table == tree.root->entries[0]->entries[0]->entries[0]->entries[0]);
    TEST_CHECK_RET(range64k.start_index == 16);
    TEST_CHECK_RET(range4k.start_index == 256);

    uvm_page_tree_put_ptes(&tree, &range64k, NULL);
    TEST_CHECK_RET(range4k.table->ref_count == 256);
    TEST_CHECK_RET(range4k.table == tree.root->entries[0]->entries[0]->entries[0]->entries[1]);
    uvm_page_tree_put_ptes(&tree, &range4k, NULL);

    UVM_ASSERT(tree.root->ref_count == 0);
    uvm_page_tree_deinit(&tree);
    return NV_OK;
}

static NV_STATUS split_and_free(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range;

    // 45 = 1 + 2 + 3 + ... + 9
    NvU64 size = 45 * (2 << 20);
    NvU32 i;
    NvU32 sum = 0;

    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, BIG_PAGE_SIZE_PASCAL, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_2M, 0, size, NULL, &range), NV_OK);
    TEST_CHECK_RET(range.entry_count == 45);
    TEST_CHECK_RET(range.table->depth == 3);
    TEST_CHECK_RET(range.start_index == 0);

    for (i = 1; i <= 9; i++) {
        range.entry_count = i;
        range.start_index = sum;
        uvm_page_tree_put_ptes(&tree, &range, NULL);
        sum += i;
    }
    UVM_ASSERT(tree.root->ref_count == 0);
    uvm_page_tree_deinit(&tree);
    return NV_OK;
}

static NV_STATUS check_sizes(uvm_gpu_t *gpu)
{
    NvU32 user_sizes = UVM_PAGE_SIZE_2M | UVM_PAGE_SIZE_64K | UVM_PAGE_SIZE_4K;
    NvU32 kernel_sizes = UVM_PAGE_SIZE_4K | 256;
    TEST_CHECK_RET(uvm_mmu_user_chunk_sizes(gpu) == user_sizes);
    TEST_CHECK_RET(uvm_mmu_kernel_chunk_sizes(gpu) == kernel_sizes);
    return NV_OK;
}

static NV_STATUS alloc_64k_memory_kepler(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range;

    NvLength size = 64 * 1024;
    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, UVM_PAGE_SIZE_64K, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_64K, 0, size, NULL, &range), NV_OK);
    TEST_CHECK_RET(range.entry_count == 1);
    TEST_CHECK_RET(range.table->depth == 1);
    TEST_CHECK_RET(range.start_index == 0);
    TEST_CHECK_RET(tree.root->ref_count == 1);
    TEST_CHECK_RET(tree.root->entries[0]->ref_count == 1);
    TEST_CHECK_RET(range.table == tree.root->entries[0]);
    uvm_page_tree_put_ptes(&tree, &range, NULL);
    UVM_ASSERT(tree.root->ref_count == 0);
    uvm_page_tree_deinit(&tree);
    return NV_OK;
}

static NV_STATUS alloc_128k_memory_kepler(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range;
    NvLength size = 128 * 1024;

    // 64k big page mode
    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, UVM_PAGE_SIZE_64K, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_64K, 0, size, NULL, &range), NV_OK);
    TEST_CHECK_RET(range.entry_count == 2);
    TEST_CHECK_RET(range.table->depth == 1);
    TEST_CHECK_RET(range.start_index == 0);
    TEST_CHECK_RET(range.page_size == UVM_PAGE_SIZE_64K);
    TEST_CHECK_RET(tree.root->ref_count == 1);
    TEST_CHECK_RET(tree.root->entries[0]->ref_count == 2);
    TEST_CHECK_RET(range.table == tree.root->entries[0]);
    uvm_page_tree_put_ptes(&tree, &range, NULL);
    UVM_ASSERT(tree.root->ref_count == 0);
    uvm_page_tree_deinit(&tree);

    // 128k big page mode
    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, UVM_PAGE_SIZE_128K, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_128K, 0, size, NULL, &range), NV_OK);
    TEST_CHECK_RET(range.entry_count == 1);
    TEST_CHECK_RET(range.table->depth == 1);
    TEST_CHECK_RET(range.start_index == 0);
    TEST_CHECK_RET(tree.root->ref_count == 1);
    TEST_CHECK_RET(range.page_size == UVM_PAGE_SIZE_128K);
    TEST_CHECK_RET(tree.root->entries[0]->ref_count == 1);
    TEST_CHECK_RET(range.table == tree.root->entries[0]);
    uvm_page_tree_put_ptes(&tree, &range, NULL);
    UVM_ASSERT(tree.root->ref_count == 0);
    uvm_page_tree_deinit(&tree);

    return NV_OK;

}

static NV_STATUS entry_test_pascal(uvm_gpu_t *gpu)
{
    NvU64 pde_bits[2];
    uvm_mmu_page_table_alloc_t *phys_allocs[2] = {NULL, NULL};
    uvm_mmu_page_table_alloc_t alloc_sys =     {.aperture = UVM_APERTURE_SYS, .addr = 0x399999999999000LL};
    uvm_mmu_page_table_alloc_t alloc_vid =     {.aperture = UVM_APERTURE_VID, .addr = 0x1BBBBBB000LL};
    // big versions have [11:8] set as well to test the page table merging
    uvm_mmu_page_table_alloc_t alloc_big_sys = {.aperture = UVM_APERTURE_SYS, .addr = 0x399999999999900LL};
    uvm_mmu_page_table_alloc_t alloc_big_vid = {.aperture = UVM_APERTURE_VID, .addr = 0x1BBBBBBB00LL};


    uvm_mmu_mode_hal_t *hal = gpu->arch_hal->mmu_mode_hal(UVM_PAGE_SIZE_64K);

    // Make sure cleared PDEs  work as expected
    hal->make_pde(pde_bits, phys_allocs, 0);
    TEST_CHECK_RET(pde_bits[0] == 0);

    memset(pde_bits, 0xFF, sizeof(pde_bits));
    hal->make_pde(pde_bits, phys_allocs, 3);
    TEST_CHECK_RET(pde_bits[0] == 0 && pde_bits[1] == 0);

    // Sys and videmem PDEs
    phys_allocs[0] = &alloc_sys;
    hal->make_pde(pde_bits, phys_allocs, 0);
    TEST_CHECK_RET(pde_bits[0] == 0x3999999999990C);

    phys_allocs[0] = &alloc_vid;
    hal->make_pde(pde_bits, phys_allocs, 0);
    TEST_CHECK_RET(pde_bits[0] == 0x1BBBBBB0A);

    // Dual PDEs
    phys_allocs[0] = &alloc_big_sys;
    phys_allocs[1] = &alloc_vid;
    hal->make_pde(pde_bits, phys_allocs, 3);
    TEST_CHECK_RET(pde_bits[0] == 0x3999999999999C && pde_bits[1] == 0x1BBBBBB0A);

    phys_allocs[0] = &alloc_big_vid;
    phys_allocs[1] = &alloc_sys;
    hal->make_pde(pde_bits, phys_allocs, 3);
    TEST_CHECK_RET(pde_bits[0] == 0x1BBBBBBBA && pde_bits[1] == 0x3999999999990C);

    // Page table entries
    TEST_CHECK_RET(hal->empty_pte(UVM_PAGE_SIZE_2M) == 0);
    TEST_CHECK_RET(hal->empty_pte(UVM_PAGE_SIZE_64K) == 0x20);
    TEST_CHECK_RET(hal->empty_pte(UVM_PAGE_SIZE_4K) == 0);

    TEST_CHECK_RET(hal->make_pte(UVM_APERTURE_SYS,
                                 0x399999999999000LL,
                                 UVM_PROT_READ_WRITE_ATOMIC,
                                 NV_TRUE, // volatile
                                 UVM_PAGE_SIZE_2M) == 0x3999999999990D);


    // change volatile to false
    TEST_CHECK_RET(hal->make_pte(UVM_APERTURE_SYS,
                                 0x399999999999000LL,
                                 UVM_PROT_READ_WRITE_ATOMIC,
                                 NV_FALSE,
                                 UVM_PAGE_SIZE_4K) == 0x39999999999905);

    // remove atomic
    TEST_CHECK_RET(hal->make_pte(UVM_APERTURE_SYS,
                                 0x399999999999000LL,
                                 UVM_PROT_READ_WRITE,
                                 NV_FALSE,
                                 UVM_PAGE_SIZE_4K) == 0x39999999999985);

    // read only
    TEST_CHECK_RET(hal->make_pte(UVM_APERTURE_SYS,
                                 0x399999999999000LL,
                                 UVM_PROT_READ_ONLY,
                                 NV_FALSE,
                                 UVM_PAGE_SIZE_4K) == 0x399999999999C5);

    // local video
    TEST_CHECK_RET(hal->make_pte(UVM_APERTURE_VID,
                                 0x1BBBBBB000LL,
                                 UVM_PROT_READ_ONLY,
                                 NV_FALSE,
                                 UVM_PAGE_SIZE_4K) == 0x1BBBBBBC1);

    // peer 0
    TEST_CHECK_RET(hal->make_pte(UVM_APERTURE_PEER_0,
                                 0x1BBBBBB000LL,
                                 UVM_PROT_READ_ONLY,
                                 NV_FALSE,
                                 UVM_PAGE_SIZE_4K) == 0x1BBBBBBC3);

    // peer 7
    TEST_CHECK_RET(hal->make_pte(UVM_APERTURE_PEER_7,
                                 0x1BBBBBB000LL,
                                 UVM_PROT_READ_ONLY,
                                 NV_FALSE,
                                 UVM_PAGE_SIZE_4K) == 0xFBBBBBBC3);

    return NV_OK;
}

static NV_STATUS entry_test_kepler(uvm_gpu_t *gpu)
{
    NvU64 pde_bits;
    uvm_mmu_page_table_alloc_t *phys_allocs[2] = {NULL, NULL};
    uvm_mmu_page_table_alloc_t alloc_sys = {.aperture = UVM_APERTURE_SYS, .addr = 0x9999999000LL};
    uvm_mmu_page_table_alloc_t alloc_vid = {.aperture = UVM_APERTURE_VID, .addr = 0x1BBBBBB000LL};

    uvm_mmu_mode_hal_t *hal = gpu->arch_hal->mmu_mode_hal(UVM_PAGE_SIZE_64K);

    hal->make_pde(&pde_bits, phys_allocs, 0);
    TEST_CHECK_RET(pde_bits == 0x0L);

    phys_allocs[0] = &alloc_sys;
    phys_allocs[1] = &alloc_vid;
    hal->make_pde(&pde_bits, phys_allocs, 0);
    TEST_CHECK_RET(pde_bits == 0x1BBBBBBD99999992LL);

    phys_allocs[0] = &alloc_vid;
    phys_allocs[1] = &alloc_sys;
    hal->make_pde(&pde_bits, phys_allocs, 0);
    TEST_CHECK_RET(pde_bits == 0x9999999E1BBBBBB1LL);

    TEST_CHECK_RET(hal->empty_pte(UVM_PAGE_SIZE_4K) == 0);

    TEST_CHECK_RET(hal->make_pte(UVM_APERTURE_SYS,
                                 0x9999999000LL,
                                 UVM_PROT_READ_WRITE_ATOMIC,
                                 NV_TRUE, // volatile
                                 UVM_PAGE_SIZE_4K) == 0x599999991LL);

    // change volatile to false
    TEST_CHECK_RET(hal->make_pte(UVM_APERTURE_SYS,
                                 0x9999999000LL,
                                 UVM_PROT_READ_WRITE_ATOMIC,
                                 NV_FALSE,
                                 UVM_PAGE_SIZE_4K) == 0x499999991LL);
    // remove atomic
    TEST_CHECK_RET(hal->make_pte(UVM_APERTURE_SYS,
                                 0x9999999000LL,
                                 UVM_PROT_READ_WRITE,
                                 NV_FALSE,
                                 UVM_PAGE_SIZE_4K) == 0x499999991LL);

    // read only
    TEST_CHECK_RET(hal->make_pte(UVM_APERTURE_SYS,
                                 0x9999999000LL,
                                 UVM_PROT_READ_ONLY,
                                 NV_FALSE,
                                 UVM_PAGE_SIZE_4K) == 0x8000000499999995LL);

    // local video
    TEST_CHECK_RET(hal->make_pte(UVM_APERTURE_VID,
                                 0x1BBBBBB000LL,
                                 UVM_PROT_READ_ONLY,
                                 NV_FALSE,
                                 UVM_PAGE_SIZE_4K) == 0x800000001BBBBBB5LL);

    // peer 0
    TEST_CHECK_RET(hal->make_pte(UVM_APERTURE_PEER_0,
                                 0x1BBBBBB000LL,
                                 UVM_PROT_READ_ONLY,
                                 NV_FALSE,
                                 UVM_PAGE_SIZE_4K) == 0x800000021BBBBBB5LL);

    // peer 7
    TEST_CHECK_RET(hal->make_pte(UVM_APERTURE_PEER_7,
                                 0x1BBBBBB000LL,
                                 UVM_PROT_READ_ONLY,
                                 NV_FALSE,
                                 UVM_PAGE_SIZE_4K) == 0x80000002FBBBBBB5LL);

    return NV_OK;
}

static NV_STATUS alloc_4k_kepler(uvm_gpu_t *gpu)
{
    uvm_page_tree_t tree;
    uvm_page_table_range_t range;
    NvLength size = 4096;

    // 64k big page mode
    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, UVM_PAGE_SIZE_64K, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_4K, 0, size, NULL, &range), NV_OK);
    TEST_CHECK_RET(range.entry_count == 1);
    TEST_CHECK_RET(range.table->depth == 1);
    TEST_CHECK_RET(range.start_index == 0);
    TEST_CHECK_RET(range.page_size == UVM_PAGE_SIZE_4K);
    TEST_CHECK_RET(tree.root->ref_count == 1);
    TEST_CHECK_RET(range.table == tree.root->entries[1]);
    TEST_CHECK_RET(tree.root->entries[1]->ref_count == 1);
    uvm_page_tree_put_ptes(&tree, &range, NULL);
    UVM_ASSERT(tree.root->ref_count == 0);
    uvm_page_tree_deinit(&tree);

    // 128k big page mode
    MEM_NV_CHECK_RET(uvm_page_tree_init(gpu, UVM_PAGE_SIZE_128K, &tree, NULL), NV_OK);
    MEM_NV_CHECK_RET(uvm_page_tree_get_ptes(&tree, UVM_PAGE_SIZE_4K, 0, size, NULL, &range), NV_OK);
    TEST_CHECK_RET(range.entry_count == 1);
    TEST_CHECK_RET(range.table->depth == 1);
    TEST_CHECK_RET(range.start_index == 0);
    TEST_CHECK_RET(range.page_size == UVM_PAGE_SIZE_4K);
    TEST_CHECK_RET(tree.root->ref_count == 1);
    TEST_CHECK_RET(range.table == tree.root->entries[1]);
    TEST_CHECK_RET(tree.root->entries[1]->ref_count == 1);
    uvm_page_tree_put_ptes(&tree, &range, NULL);
    UVM_ASSERT(tree.root->ref_count == 0);
    uvm_page_tree_deinit(&tree);

    return NV_OK;
}

static uvm_ce_hal_t fake_ce_hal = { .memset_8 = fake_ce_memset_8 };

static NV_STATUS fake_gpu_init(NvU32 host_class, NvU32 ce_class, NvU32 architecture, uvm_gpu_t *fake_gpu)
{
    memset(fake_gpu, 0, sizeof(*fake_gpu));

    fake_gpu->ce_class = ce_class;
    fake_gpu->host_class = host_class;
    fake_gpu->architecture = architecture;

    TEST_CHECK_RET(uvm_hal_init_gpu(fake_gpu) == NV_OK);

    fake_gpu->ce_hal = &fake_ce_hal;

    return NV_OK;
}

static NV_STATUS fake_gpu_init_kepler(uvm_gpu_t *fake_gpu)
{
    return fake_gpu_init(KEPLER_CHANNEL_GPFIFO_A, KEPLER_DMA_COPY_A, NV2080_CTRL_MC_ARCH_INFO_ARCHITECTURE_GK100, fake_gpu);
}

static NV_STATUS fake_gpu_init_pascal(uvm_gpu_t *fake_gpu)
{
    return fake_gpu_init(PASCAL_CHANNEL_GPFIFO_A, PASCAL_DMA_COPY_A, NV2080_CTRL_MC_ARCH_INFO_ARCHITECTURE_GP100, fake_gpu);
}

static NV_STATUS kepler_test_page_tree(void)
{
    // create a fake Kepler GPU for this test.
    uvm_gpu_t kepler;

    TEST_CHECK_RET(fake_gpu_init_kepler(&kepler) == NV_OK);

    MEM_NV_CHECK_RET(allocate_root(&kepler), NV_OK);
    MEM_NV_CHECK_RET(alloc_64k_memory_kepler(&kepler), NV_OK);
    MEM_NV_CHECK_RET(alloc_128k_memory_kepler(&kepler), NV_OK);
    MEM_NV_CHECK_RET(alloc_4k_kepler(&kepler), NV_OK);
    TEST_CHECK_RET(entry_test_kepler(&kepler) == NV_OK);
    return NV_OK;
}

NV_STATUS uvm8_test_page_tree(UVM_TEST_PAGE_TREE_PARAMS *params, struct file *filp)
{
    // create a fake Pascal GPU for this test.
    uvm_gpu_t pascal;

    TEST_CHECK_RET(fake_gpu_init_pascal(&pascal) == NV_OK);

    MEM_NV_CHECK_RET(allocate_root(&pascal), NV_OK);
    MEM_NV_CHECK_RET(alloc_64k_memory(&pascal), NV_OK);
    MEM_NV_CHECK_RET(alloc_adjacent_64k_memory(&pascal), NV_OK);
    MEM_NV_CHECK_RET(alloc_adjacent_pde_64k_memory(&pascal), NV_OK);
    MEM_NV_CHECK_RET(alloc_nearby_pde_64k_memory(&pascal), NV_OK);
    MEM_NV_CHECK_RET(allocate_then_free_all_16_64k(&pascal), NV_OK);
    MEM_NV_CHECK_RET(allocate_then_free_8_8_64k(&pascal), NV_OK);
    MEM_NV_CHECK_RET(get_single_page_2m(&pascal), NV_OK);
    MEM_NV_CHECK_RET(get_entire_table_4k(&pascal), NV_OK);
    MEM_NV_CHECK_RET(split_4k_from_2m(&pascal), NV_OK);
    MEM_NV_CHECK_RET(get_512mb_range(&pascal), NV_OK);
    MEM_NV_CHECK_RET(get_two_free_apart(&pascal), NV_OK);
    MEM_NV_CHECK_RET(get_overlapping_dual_pdes(&pascal), NV_OK);
    MEM_NV_CHECK_RET(split_and_free(&pascal), NV_OK);
    MEM_NV_CHECK_RET(entry_test_pascal(&pascal), NV_OK);
    MEM_NV_CHECK_RET(check_sizes(&pascal), NV_OK);
    MEM_NV_CHECK_RET(kepler_test_page_tree(), NV_OK);
    return NV_OK;
}
