/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#ifndef __UVM8_PERF_UTILS_H__
#define __UVM8_PERF_UTILS_H__

#include "linux/log2.h"

#include "nvstatus.h"
#include "nvtypes.h"

// Macros to perform increments that saturate at the maximum values allowed by the variable underlying storage
#define UVM_PERF_SATURATING_ADD(counter,value)     \
({                                                 \
    NvU64 expected;                                \
    NvU64 old;                                     \
                                                   \
    old = counter;                                 \
    expected = (NvU64)counter + (NvU64)value;      \
    counter += value;                              \
    if (counter != expected || expected < old)     \
        counter = -1;                              \
    counter;                                       \
})

#define UVM_PERF_SATURATING_INC(counter) UVM_PERF_SATURATING_ADD(counter, 1)

// Array-based tree type for fix-sized binary trees. Nodes are stored in a contiguous array, ordered per level (from
// the leaf to the root). These trees are mainly used to keep statistics for memory regions. Stats are updated from a
// leaf node (which typically represents a page) up to the root of the tree (which represents the whole memory region
// tracked by the tree). Thus, statistics are transparently aggregated for different memory region size granularities.
//
// Restrictions: trees of up 64 levels are supported for 64-bit architectures, 32 levels for 32-bit architectures

typedef struct
{
    // Number of leaves
    NvLength leaf_count;

    // Number of node levels in the tree (32/64 maximum, so we can use a single byte)
    NvU8 level_count;

    // Total number of nodes (leaf + internal)
    NvLength node_count;

    void *nodes;
} uvm_perf_tree_t;

// Tree traversal
//
// Full- and complete-binary trees have properties that enable easy traversal of the tree using simple division and
// multiplication operations. However, forcing trees to be complete could leave to a huge waste of memory (up to 25%
// per tree). Therefore, tree traversals here need to compute the number of elements in the current/lower level before
// moving to the upper/lower level. Level 0 is the root level, (level_count - 1) is the leaves' level. This can be
// easily done with the following computation (see uvm_perf_tree_level_node_count):
//
// (1 << level) - (missing_leaves_to_pow2 >> ((levels - 1) - level))
//
// Once we have the offset to the beginning of the current level, we only need to add the index of the node visited
// within the current level. This is done as follows (see uvm_perf_tree_level_node_index):
//
// leaf_idx >> ((levels - 1) - level)
//
// We provide a type for tree traversals to allow macros to store the necessary information to transparently perform
// these computations. Thus, an uvm_perf_tree_iter_t object needs to be passed to the tree traversal macros.
typedef struct
{
    // Leaf index used to start the tree traversal
    NvLength leaf_idx;

    // Maximum valid level index in the traversed tree
    NvS8 max_level_idx;

    // Current level in the traversal. Needs to be negative to allow detecting when we are out of bounds
    NvS8 level_idx;

    // Offset of the current level within the node array
    NvLength level_offset;

    // Number of leaf nodes that would make this a full binary tree
    NvLength missing_leaves_to_pow2;
} uvm_perf_tree_iter_t;

// Tree initialization. Computes the total number of levels and nodes required for the given number of leaf nodes and
// allocates its memory. Nodes' memory is zero-initialized.
//
// Returns NV_OK if initialization succeeded, or
// NV_ERR_NO_MEMORY if node allocation failed
NV_STATUS uvm_perf_tree_init(uvm_perf_tree_t *tree, size_t node_size, NvLength leaf_count);

// Tree destruction. It frees the memory used by the nodes
void uvm_perf_tree_destroy(uvm_perf_tree_t *tree);

// Initializes the context for a tree traversal from the given leaf to the root of the tree
static void uvm_perf_tree_init_up_traversal(uvm_perf_tree_t *tree, NvLength leaf, uvm_perf_tree_iter_t *iter)
{
   NvLength pow2_leaf_count;

   pow2_leaf_count = roundup_pow_of_two(tree->leaf_count);

   iter->leaf_idx      = leaf;

   iter->max_level_idx = tree->level_count - 1;
   iter->level_idx     = tree->level_count - 1;
   iter->level_offset  = 0;

   iter->missing_leaves_to_pow2 = pow2_leaf_count - tree->leaf_count;
}

// Initializes the context for a tree traversal from the given leaf to the root of the tree
static void uvm_perf_tree_init_down_traversal(uvm_perf_tree_t *tree, NvLength leaf, uvm_perf_tree_iter_t *iter)
{
   NvLength pow2_leaf_count;

   pow2_leaf_count = roundup_pow_of_two(tree->leaf_count);

   iter->leaf_idx      = leaf;

   iter->max_level_idx = tree->level_count - 1;
   iter->level_idx     = 0;
   iter->level_offset  = tree->node_count - 1;

   iter->missing_leaves_to_pow2 = pow2_leaf_count - tree->leaf_count;
}

// Computes the index of the node visited for the traversal in the current level
static NvLength uvm_perf_tree_level_node_index(uvm_perf_tree_iter_t *iter)
{
    return iter->leaf_idx >> (iter->max_level_idx - iter->level_idx);
}

// Computes the number of nodes in the given level
static NvLength uvm_perf_tree_level_node_count(uvm_perf_tree_iter_t *iter, NvLength level_idx)
{
    NvLength level_pow2_node_count;
    NvLength level_missing_nodes_to_pow2;

    level_pow2_node_count       = 1 << level_idx;
    level_missing_nodes_to_pow2 = iter->missing_leaves_to_pow2 >> (iter->max_level_idx - level_idx);

    return level_pow2_node_count - level_missing_nodes_to_pow2;
}

// Obtains the current node pointed by the traversal context
#define UVM_PERF_TREE_ITER_CURRENT(tree,node_type,iter)                                             \
({                                                                                                  \
    (iter)->level_idx < 0 || (iter)->level_idx > (iter)->max_level_idx ?                            \
        NULL:                                                                                       \
        ((node_type *)(tree)->nodes) + (iter)->level_offset + uvm_perf_tree_level_node_index(iter); \
})

// Obtains the leaf node corresponding to the given leaf node index
#define UVM_PERF_TREE_LEAF(tree,node_type,leaf_idx) \
({                                                  \
    ((node_type *)(tree)->nodes) + leaf_idx;        \
})

// Obtains the root node of the tree
#define UVM_PERF_TREE_ROOT(tree,node_type)                   \
({                                                           \
    ((node_type *)(tree)->nodes) + ((tree)->node_count - 1); \
})

// Functions to updates the tree traversal context with the information of the next level (up/down)
static void uvm_perf_tree_traverse_up(uvm_perf_tree_iter_t *iter)
{
    // Nodes of the next level (up) are stored AFTER the current level
    iter->level_offset += uvm_perf_tree_level_node_count(iter, iter->level_idx);

    --iter->level_idx;
}

static void uvm_perf_tree_traverse_down(uvm_perf_tree_iter_t *iter)
{
    // Nodes of the next level (down) are stored BEFORE the current level. Since we are at the beginning of the current
    // level, we must skip all the nodes of the NEXT level.
    iter->level_offset -= uvm_perf_tree_level_node_count(iter, iter->level_idx + 1);

    ++iter->level_idx;
}

// Complete tree traversal from the given leaf up to the root of the tree. A pointer to the node in each level of the
// traversal is stored in node
#define uvm_perf_tree_traverse_leaf_to_root(tree,leaf,node,iter)           \
    for (uvm_perf_tree_init_up_traversal(tree, leaf, iter),                \
         (node) = UVM_PERF_TREE_ITER_CURRENT(tree, typeof(*(node)), iter); \
         (node) != NULL;                                                   \
         uvm_perf_tree_traverse_up(iter),                                  \
         (node) = UVM_PERF_TREE_ITER_CURRENT(tree, typeof(*(node)), iter))

// Complete tree traversal from the root of the tree down to the given leaf index. A pointer to the node in each level
// of the traversal is stored in node
#define uvm_perf_tree_traverse_root_to_leaf(tree,leaf,node,iter)           \
    for (uvm_perf_tree_init_down_traversal(tree, leaf, iter),              \
         (node) = UVM_PERF_TREE_ITER_CURRENT(tree, typeof(*(node)), iter); \
         (node) != NULL;                                                   \
         uvm_perf_tree_traverse_down(iter),                                \
         (node) = UVM_PERF_TREE_ITER_CURRENT(tree, typeof(*(node)), iter))

#endif
