/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm8_perf_utils.h"
#include "uvm8_test.h"

static NV_STATUS test_saturating_counter_basic(void)
{
    NvU8 counter8;
    NvU16 counter16;
    NvU32 counter32;
    NvU64 counter64;

    NvU8 max8   = -1;
    NvU16 max16 = -1;
    NvU32 max32 = -1;
    NvU64 max64 = -1;

    counter8 = 0;

    UVM_PERF_SATURATING_INC(counter8);
    TEST_CHECK_RET(counter8 == 1);
    UVM_PERF_SATURATING_INC(counter8);
    TEST_CHECK_RET(counter8 == 2);

    UVM_PERF_SATURATING_ADD(counter8, 5);
    TEST_CHECK_RET(counter8 == 7);

    // Counter saturating at maximum unsigned char value
    UVM_PERF_SATURATING_ADD(counter8, max8);
    TEST_CHECK_RET(counter8 == max8);

    counter16 = 0;

    UVM_PERF_SATURATING_INC(counter16);
    TEST_CHECK_RET(counter16 == 1);
    UVM_PERF_SATURATING_INC(counter16);
    TEST_CHECK_RET(counter16 == 2);

    UVM_PERF_SATURATING_ADD(counter16, 5);
    TEST_CHECK_RET(counter16 == 7);

    // Counter saturating at maximum unsigned short value
    UVM_PERF_SATURATING_ADD(counter16, max16);
    TEST_CHECK_RET(counter16 == max16);

    counter32 = 0;

    UVM_PERF_SATURATING_INC(counter32);
    TEST_CHECK_RET(counter32 == 1);
    UVM_PERF_SATURATING_INC(counter32);
    TEST_CHECK_RET(counter32 == 2);

    UVM_PERF_SATURATING_ADD(counter32, 5);
    TEST_CHECK_RET(counter32 == 7);

    // Counter saturating at maximum unsigned long int value
    UVM_PERF_SATURATING_ADD(counter32, max32);
    TEST_CHECK_RET(counter32 == max32);

    counter64 = 0;

    UVM_PERF_SATURATING_INC(counter64);
    TEST_CHECK_RET(counter64 == 1);
    UVM_PERF_SATURATING_INC(counter64);
    TEST_CHECK_RET(counter64 == 2);

    UVM_PERF_SATURATING_ADD(counter64, 5);
    TEST_CHECK_RET(counter64 == 7);

    // Counter saturating at maximum unsigned long long int value
    UVM_PERF_SATURATING_ADD(counter64, max64);
    TEST_CHECK_RET(counter64 == max64);

    return NV_OK;
}

struct region
{
    unsigned char   read_faults : 4;
    unsigned char  write_faults : 4;
    unsigned char atomic_faults : 4;
    unsigned char      upgrades : 4;
};

static NV_STATUS test_saturating_counter_bitfields(void)
{
    struct region r;

    memset(&r, 0, sizeof(r));

    UVM_PERF_SATURATING_INC(r.read_faults);
    TEST_CHECK_RET(r.read_faults == 1);
    UVM_PERF_SATURATING_INC(r.write_faults);
    TEST_CHECK_RET(r.write_faults == 1);
    UVM_PERF_SATURATING_ADD(r.atomic_faults, 18);
    TEST_CHECK_RET(r.atomic_faults == 15);

    return NV_OK;
}

static NV_STATUS test_saturating_counter(void)
{
    NV_STATUS status;

    status = test_saturating_counter_basic();
    if (status != NV_OK)
        goto fail;
    status = test_saturating_counter_bitfields();

fail:
    return status;
}

static NV_STATUS test_tree_pow2(void)
{
    NV_STATUS status;
    uvm_perf_tree_t my_int_tree;
    uvm_perf_tree_iter_t ctx;
    int *node;

    status = uvm_perf_tree_init(&my_int_tree, sizeof(int), 8);
    MEM_NV_CHECK_RET(status, NV_OK);

    TEST_CHECK_RET(my_int_tree.node_count == 8 * 2 - 1);
    TEST_CHECK_RET(my_int_tree.level_count == 4);

    uvm_perf_tree_traverse_leaf_to_root(&my_int_tree, 3, node, &ctx) {
        ++*node;
    }

    // Level 0 (leafs)
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[1] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[2] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[3] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[4] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[5] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[6] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[7] == 0);
    // Level 1
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8 + 0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8 + 1] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8 + 2] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8 + 3] == 0);
    // Level 2
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[12 + 0] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[12 + 1] == 0);
    // Level 3 (root)
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 0] == 1);

    TEST_CHECK_RET(*UVM_PERF_TREE_ROOT(&my_int_tree, int) == 1);

    uvm_perf_tree_traverse_leaf_to_root(&my_int_tree, 6, node, &ctx) {
        ++*node;
    }
    // Level 0 (leafs)
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[1] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[2] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[3] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[4] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[5] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[6] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[7] == 0);
    // Level 1
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8 + 0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8 + 1] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8 + 2] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8 + 3] == 1);
    // Level 2
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[12 + 0] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[12 + 1] == 1);
    // Level 3 (root)
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 0] == 2);

    TEST_CHECK_RET(*UVM_PERF_TREE_ROOT(&my_int_tree, int) == 2);

    uvm_perf_tree_traverse_root_to_leaf(&my_int_tree, 7, node, &ctx) {
        ++*node;
    }

    // Level 0 (leafs)
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[1] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[2] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[3] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[4] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[5] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[6] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[7] == 1);

    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 0) == &((int *)my_int_tree.nodes)[0]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 1) == &((int *)my_int_tree.nodes)[1]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 2) == &((int *)my_int_tree.nodes)[2]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 3) == &((int *)my_int_tree.nodes)[3]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 4) == &((int *)my_int_tree.nodes)[4]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 5) == &((int *)my_int_tree.nodes)[5]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 6) == &((int *)my_int_tree.nodes)[6]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 7) == &((int *)my_int_tree.nodes)[7]);

    // Level 1
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8 + 0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8 + 1] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8 + 2] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8 + 3] == 2);
    // Level 2
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[12 + 0] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[12 + 1] == 2);
    // Level 3 (root)
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 0] == 3);

    TEST_CHECK_RET(UVM_PERF_TREE_ROOT(&my_int_tree, int) == &((int *)my_int_tree.nodes)[14 + 0]);

    uvm_perf_tree_destroy(&my_int_tree);

    return NV_OK;
}

static NV_STATUS test_tree_non_pow2(void)
{
    NV_STATUS status;
    uvm_perf_tree_t my_int_tree;
    uvm_perf_tree_iter_t ctx;
    int *node;

    status = uvm_perf_tree_init(&my_int_tree, sizeof(int), 7);
    MEM_NV_CHECK_RET(status, NV_OK);

    TEST_CHECK_RET(my_int_tree.node_count == 7 + 4 + 2 + 1);
    TEST_CHECK_RET(my_int_tree.level_count == 4);

    uvm_perf_tree_destroy(&my_int_tree);

    status = uvm_perf_tree_init(&my_int_tree, sizeof(int), 9);
    MEM_NV_CHECK_RET(status, NV_OK);

    TEST_CHECK_RET(my_int_tree.node_count == 9 + 5 + 3 + 2 + 1);
    TEST_CHECK_RET(my_int_tree.level_count == 5);

    uvm_perf_tree_destroy(&my_int_tree);

    status = uvm_perf_tree_init(&my_int_tree, sizeof(int), 13);
    MEM_NV_CHECK_RET(status, NV_OK);

    TEST_CHECK_RET(my_int_tree.node_count == 13 + 7 + 4 + 2 + 1);
    TEST_CHECK_RET(my_int_tree.level_count == 5);

    uvm_perf_tree_destroy(&my_int_tree);

    status = uvm_perf_tree_init(&my_int_tree, sizeof(int), 15);
    MEM_NV_CHECK_RET(status, NV_OK);

    TEST_CHECK_RET(my_int_tree.node_count == 15 + 8 + 4 + 2 + 1);
    TEST_CHECK_RET(my_int_tree.level_count == 5);

    uvm_perf_tree_destroy(&my_int_tree);

    status = uvm_perf_tree_init(&my_int_tree, sizeof(int), 9);
    MEM_NV_CHECK_RET(status, NV_OK);

    uvm_perf_tree_traverse_leaf_to_root(&my_int_tree, 6, node, &ctx) {
        ++*node;
    }
    // Level 0 (leafs)
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[1] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[2] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[3] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[4] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[5] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[6] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[7] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8] == 0);
    // Level 1
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 1] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 2] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 3] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 4] == 0);
    // Level 2
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 1] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 2] == 0);
    // Level 3
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[17 + 0] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[17 + 1] == 0);
    // Level 4 (root)
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[19 + 0] == 1);

    TEST_CHECK_RET(*UVM_PERF_TREE_ROOT(&my_int_tree, int) == 1);

    uvm_perf_tree_traverse_root_to_leaf(&my_int_tree, 2, node, &ctx) {
        ++*node;
    }

    uvm_perf_tree_traverse_root_to_leaf(&my_int_tree, 8, node, &ctx) {
        ++*node;
    }

    // Level 0 (leafs)
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[1] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[2] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[3] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[4] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[5] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[6] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[7] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8] == 1);

    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 0) == &((int *)my_int_tree.nodes)[0]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 1) == &((int *)my_int_tree.nodes)[1]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 2) == &((int *)my_int_tree.nodes)[2]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 3) == &((int *)my_int_tree.nodes)[3]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 4) == &((int *)my_int_tree.nodes)[4]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 5) == &((int *)my_int_tree.nodes)[5]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 6) == &((int *)my_int_tree.nodes)[6]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 7) == &((int *)my_int_tree.nodes)[7]);
    TEST_CHECK_RET(UVM_PERF_TREE_LEAF(&my_int_tree, int, 8) == &((int *)my_int_tree.nodes)[8]);

    // Level 1
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 1] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 2] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 3] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 4] == 1);
    // Level 2
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 0] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 1] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 2] == 1);
    // Level 3
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[17 + 0] == 2);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[17 + 1] == 1);
    // Level 4 (root)
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[19 + 0] == 3);

    TEST_CHECK_RET(UVM_PERF_TREE_ROOT(&my_int_tree, int) == &((int *)my_int_tree.nodes)[19 + 0]);

    uvm_perf_tree_destroy(&my_int_tree);

    return NV_OK;
}

static NV_STATUS test_tree_traversal(void)
{
    NV_STATUS status;
    uvm_perf_tree_t my_int_tree;
    uvm_perf_tree_iter_t ctx;
    int *node;
    int value;

    status = uvm_perf_tree_init(&my_int_tree, sizeof(int), 9);
    MEM_NV_CHECK_RET(status, NV_OK);

    value = 1;

    uvm_perf_tree_traverse_leaf_to_root(&my_int_tree, 6, node, &ctx) {
        *node += value++;
    }
    // Level 0 (leafs)
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[1] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[2] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[3] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[4] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[5] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[6] == 1);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[7] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8] == 0);
    // Level 1
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 1] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 2] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 3] == 2);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 4] == 0);
    // Level 2
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 1] == 3);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 2] == 0);
    // Level 3
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[17 + 0] == 4);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[17 + 1] == 0);
    // Level 4 (root)
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[19 + 0] == 5);

    TEST_CHECK_RET(*UVM_PERF_TREE_ROOT(&my_int_tree, int) == 5);

    uvm_perf_tree_traverse_root_to_leaf(&my_int_tree, 6, node, &ctx) {
        *node -= --value;
    }

    // Level 0 (leafs)
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[1] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[2] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[3] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[4] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[5] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[6] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[7] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[8] == 0);
    // Level 1
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 1] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 2] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 3] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[9 + 4] == 0);
    // Level 2
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 1] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[14 + 2] == 0);
    // Level 3
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[17 + 0] == 0);
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[17 + 1] == 0);
    // Level 4 (root)
    TEST_CHECK_RET(((int *)my_int_tree.nodes)[19 + 0] == 0);

    uvm_perf_tree_destroy(&my_int_tree);

    return NV_OK;
}



static NV_STATUS test_trees(void)
{
    NV_STATUS status;

    status = test_tree_pow2();
    if (status != NV_OK)
        goto fail;
    status = test_tree_non_pow2();
    if (status != NV_OK)
        goto fail;
    status = test_tree_traversal();

fail:
    return status;
}

NV_STATUS uvm8_test_perf_utils_sanity(UVM_TEST_PERF_UTILS_SANITY_PARAMS *params, struct file *filp)
{
    NV_STATUS status;

    status = test_saturating_counter();
    if (status != NV_OK)
        goto fail;
    status = test_trees();

fail:
    return status;
}
