/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_ioctl.h"
#include "uvm8_va_space.h"
#include "uvm8_va_range.h"

static bool preferred_location_is_va_range_split_needed(uvm_va_range_t *va_range, void *data)
{
    uvm_processor_id_t processor_id;
    UVM_ASSERT(data);
    processor_id = *(uvm_processor_id_t*)data;
    return (processor_id != va_range->preferred_location);
}

static NV_STATUS preferred_location_set(uvm_va_space_t *va_space,
                                        NvU64 base,
                                        NvU64 length,
                                        uvm_processor_id_t preferred_location,
                                        uvm_va_range_t **first_va_range_to_migrate)
{
    uvm_va_range_t *va_range;
    NvU64 current_base = base;
    NvU64 last_address = base + length - 1;
    NV_STATUS status = NV_OK;

    uvm_assert_rwsem_locked_write(&va_space->lock);

    if (first_va_range_to_migrate)
        *first_va_range_to_migrate = NULL;

    status = uvm_va_space_split_span_as_needed(va_space,
                                               current_base,
                                               last_address + 1,
                                               preferred_location_is_va_range_split_needed,
                                               &preferred_location);
    if (status != NV_OK)
        return status;

    uvm_for_each_va_range_in(va_range, va_space, base, last_address) {
        // Check for potential holes in the VA ranges span
        if (va_range->node.start != current_base)
            return NV_ERR_INVALID_ADDRESS;

        status = uvm_va_range_set_preferred_location(va_range, preferred_location);
        if (status != NV_OK)
            return status;

        // Return the first VA range that needs to be migrated so the caller
        // function doesn't need to traverse the tree again
        if (first_va_range_to_migrate && !(*first_va_range_to_migrate) &&
            preferred_location != UVM8_MAX_PROCESSORS &&
            va_range->range_group && !atomic_read(&va_range->range_group->allow_migration))
            *first_va_range_to_migrate = va_range;

        current_base = va_range->node.end + 1;
    }

    // Check for potential hole at the end
    if (current_base != last_address + 1)
        return NV_ERR_INVALID_ADDRESS;

    return NV_OK;
}

NV_STATUS uvm_api_set_preferred_location(UVM_SET_PREFERRED_LOCATION_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    uvm_va_range_t *va_range = NULL;
    uvm_va_range_t *first_va_range_to_migrate = NULL;
    uvm_processor_id_t preferred_location_id;
    NV_STATUS status = NV_OK;
    bool has_va_space_write_lock;

    UVM_ASSERT(va_space);

    // Check address and length alignment
    if (!PAGE_ALIGNED(params->requestedBase) || !PAGE_ALIGNED(params->length))
        return NV_ERR_INVALID_ADDRESS;

    uvm_down_read(&g_uvm_global.global_lock);
    uvm_down_write(&va_space->lock);
    has_va_space_write_lock = true;

    // If the CPU is the preferred location, we don't have to find the associated uvm_gpu_t
    if (uvm_uuid_is_cpu(&params->preferredLocation)) {
        preferred_location_id = UVM_CPU_ID;
    }
    else {
        // Translate preferredLocation into a live GPU ID:
        status = uvm_registered_uuid_to_id(va_space, &params->preferredLocation, &preferred_location_id);
        if (status != NV_OK)
            goto done;
    }

    status = preferred_location_set(va_space, params->requestedBase, params->length, preferred_location_id,
                                    &first_va_range_to_migrate);
    if (status != NV_OK)
        goto done;

    // No VA range to migrate, early exit
    if (!first_va_range_to_migrate)
        goto done;

    uvm_downgrade_write(&va_space->lock);
    has_va_space_write_lock = false;

    // No need to check for holes in the VA ranges span here, this was checked by preferred_location_set
    for (va_range = first_va_range_to_migrate; va_range;
         va_range = uvm_va_space_iter_next(va_range, params->requestedBase + params->length - 1)) {
        if (va_range->range_group && !atomic_read(&va_range->range_group->allow_migration)) {
            // TODO: migrate VA range to its preferred location
            // TODO: update range_groups_migration test to remove manual unmaps
        }
    }

done:
    if (has_va_space_write_lock)
        uvm_up_write(&va_space->lock);
    else
        uvm_up_read(&va_space->lock);
    uvm_up_read(&g_uvm_global.global_lock);
    return status;
}

NV_STATUS uvm_api_unset_preferred_location(UVM_UNSET_PREFERRED_LOCATION_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    NV_STATUS status = NV_OK;

    UVM_ASSERT(va_space);

    uvm_down_write(&va_space->lock);

    // Check address and length alignment
    if (!PAGE_ALIGNED(params->requestedBase) || !PAGE_ALIGNED(params->length)) {
         status = NV_ERR_INVALID_ADDRESS;
         goto done;
    }

    status = preferred_location_set(va_space, params->requestedBase, params->length, UVM8_MAX_PROCESSORS, NULL);

done:
    uvm_up_write(&va_space->lock);
    return status;
}

typedef struct
{
    uvm_processor_id_t processor_id;
    bool set_bit;
} accessed_by_split_params_t;

static bool accessed_by_is_va_range_split_needed(uvm_va_range_t *va_range, void *data)
{
    accessed_by_split_params_t *params = (accessed_by_split_params_t*)data;
    UVM_ASSERT(params);

    return (uvm_processor_mask_test(&va_range->accessed_by, params->processor_id) != params->set_bit);
}

static NV_STATUS accessed_by_set(uvm_va_space_t *va_space,
                                 NvU64 base,
                                 NvU64 length,
                                 NvProcessorUuid *processor_uuid,
                                 bool set_bit)
{
    uvm_processor_id_t processor_id;
    uvm_va_range_t *va_range;
    NvU64 current_base = base;
    NvU64 last_address = base + length - 1;
    accessed_by_split_params_t split_params;
    NV_STATUS status = NV_OK;

    UVM_ASSERT(va_space);

    // Check address and length alignment
    if (!PAGE_ALIGNED(base) || !PAGE_ALIGNED(length))
        return NV_ERR_INVALID_ADDRESS;

    uvm_down_read(&g_uvm_global.global_lock);
    uvm_down_write(&va_space->lock);

    // If the CPU is the accessed by processor, we don't have to find the associated uvm_gpu_t
    if (uvm_uuid_is_cpu(processor_uuid)) {
        processor_id = UVM_CPU_ID;
    }
    else {
        // Translate processor_uuid into a live GPU ID:
        status = uvm_registered_uuid_to_id(va_space, processor_uuid, &processor_id);
        if (status != NV_OK)
            goto done;
    }

    split_params.processor_id = processor_id;
    split_params.set_bit = set_bit;
    status = uvm_va_space_split_span_as_needed(va_space,
                                               current_base,
                                               last_address + 1,
                                               accessed_by_is_va_range_split_needed,
                                               &split_params);
    if (status != NV_OK)
        goto done;

    uvm_for_each_va_range_in(va_range, va_space, base, last_address) {
        // Check for potential holes in the VA ranges span
        if (va_range->node.start != current_base) {
            status = NV_ERR_INVALID_ADDRESS;
            goto done;
        }

        if (set_bit) {
            uvm_processor_mask_set(&va_range->accessed_by, processor_id);

            // TODO: Add mappings of this VA range on that processor
        }
        else {
            uvm_processor_mask_clear(&va_range->accessed_by, processor_id);
        }

        current_base = va_range->node.end + 1;
    }

    // Check for potential hole at the end
    if (current_base != last_address + 1) {
        status = NV_ERR_INVALID_ADDRESS;
    }

done:
    uvm_up_write(&va_space->lock);
    uvm_up_read(&g_uvm_global.global_lock);
    return status;
}

NV_STATUS uvm_api_set_accessed_by(UVM_SET_ACCESSED_BY_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);

    return accessed_by_set(va_space, params->requestedBase, params->length, &params->accessedByUuid, true);
}

NV_STATUS uvm_api_unset_accessed_by(UVM_UNSET_ACCESSED_BY_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);

    return accessed_by_set(va_space, params->requestedBase, params->length, &params->accessedByUuid, false);
}

static bool read_duplication_is_va_range_split_needed(uvm_va_range_t *va_range, void *data)
{
    bool enabled;
    UVM_ASSERT(data);
    enabled = *(bool*)data;

    // Always split if the VA range read duplication policy is unset
    if (va_range->read_duplication == UVM_READ_DUPLICATION_UNSET)
        return true;

    // Otherwise, check if the new policy is different
    return (enabled != (va_range->read_duplication == UVM_READ_DUPLICATION_ENABLED));
}

static NV_STATUS read_duplication_set(uvm_va_space_t *va_space,
                                      NvU64 base,
                                      NvU64 length,
                                      bool enable)
{
    uvm_va_range_t *va_range;
    NvU64 current_base = base;
    NvU64 last_address = base + length - 1;
    NV_STATUS status = NV_OK;

    UVM_ASSERT(va_space);

    // Check address and length alignment
    if (!PAGE_ALIGNED(base) || !PAGE_ALIGNED(length))
        return NV_ERR_INVALID_ADDRESS;

    uvm_down_write(&va_space->lock);

    status = uvm_va_space_split_span_as_needed(va_space,
                                               current_base,
                                               last_address + 1,
                                               read_duplication_is_va_range_split_needed,
                                               &enable);
    if (status != NV_OK)
        goto done;

    uvm_for_each_va_range_in(va_range, va_space, base, last_address) {
        // Check for potential holes in the VA ranges span
        if (va_range->node.start != current_base) {
            status = NV_ERR_INVALID_ADDRESS;
            goto done;
        }

        va_range->read_duplication = enable ? UVM_READ_DUPLICATION_ENABLED : UVM_READ_DUPLICATION_DISABLED;

        // Expected start of next VA range
        current_base = va_range->node.end + 1;
    }

    // Check for potential hole at the end
    if (current_base != last_address + 1)
        status = NV_ERR_INVALID_ADDRESS;

done:
    uvm_up_write(&va_space->lock);
    return status;
}

NV_STATUS uvm_api_enable_read_duplication(UVM_ENABLE_READ_DUPLICATION_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);

    return read_duplication_set(va_space, params->requestedBase, params->length, true);
}

NV_STATUS uvm_api_disable_read_duplication(UVM_DISABLE_READ_DUPLICATION_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);

    return read_duplication_set(va_space, params->requestedBase, params->length, false);
}
