/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#ifndef __UVM8_PUSH_H__
#define __UVM8_PUSH_H__

#include "uvm8_forward_decl.h"
#include "uvm8_channel.h"
#include "uvm8_push_macros.h"
#include "uvm8_tracker.h"
#include "nvtypes.h"

// Space uvm_push_end() takes in the pushbuffer.
// Currently uvm_push_end() does ce_hal->semaphore_release() that takes (4 + 2)
// * 4 space. This is mostly useful for tests and there is also a test that
// verifies that it's correct.
#define UVM_PUSH_END_SIZE 24

typedef enum
{
    // By default all CE transfers are not pipelined.
    // This flag indicates that next CE transfer should be pipelined
    UVM_PUSH_FLAG_CE_NEXT_PIPELINED,

    // By default all CE operations include a membar sys after any transfer and
    // before a semaphore operation.
    // This flag indicates that next operation should use no membar at all.
    UVM_PUSH_FLAG_CE_NEXT_MEMBAR_NONE,

    // By default all CE operations include a membar sys after any transfer and
    // before a semaphore operation.
    // This flag indicates that next operation should use a membar gpu instead.
    UVM_PUSH_FLAG_CE_NEXT_MEMBAR_GPU,

    UVM_PUSH_FLAG_COUNT,
} uvm_push_flag_t;

struct uvm_push_struct
{
    // Location of the first method of the push
    NvU32 *begin;

    // Location of the next method to be written
    NvU32 *next;

    // The channel the push is being done on or has been finished on
    uvm_channel_t *channel;

    // The tracking value when the push completes on the GPU on the channel
    // above. It will be 0 for an on-going push.
    NvU64 channel_tracking_value;

    // Index for the push info stored within the channel.
    // Only valid for an on-going push (after uvm_push_begin*(), but before
    // uvm_push_end()).
    NvU32 push_info_index;

    // A bitmap of flags from uvm_push_flag_t
    DECLARE_BITMAP(flags, UVM_PUSH_FLAG_COUNT);
};

struct uvm_push_info_struct
{
    // Filename where the push was started
    const char *filename;

    // Line number where the push was started
    int line;

    // Function where the push was started
    const char *function;

    // Description of the push created from the uvm_push_begin*() format and arguments.
    char description[128];
};

// Internal implementation used by uvm_push_begin() and uvm_push_begin_acquire()
NV_STATUS __uvm_push_begin_acquire(uvm_channel_manager_t *manager, uvm_channel_type_t type, uvm_tracker_t *deps, uvm_push_t *push);

// Internal implementation used by uvm_push_begin_on_channel()
NV_STATUS __uvm_push_begin_on_channel(uvm_channel_t *channel, uvm_push_t *push);

// Is tracking push descriptions enabled?
bool uvm_push_info_is_tracking_descriptions(void);

// Internal helper to fill info push info as part of beginning a push.
void __uvm_push_fill_info(uvm_push_t *push, const char *filename, const char *function, int line, const char *format, va_list args);

// Internal helper for uvm_push_begin() and uvm_push_begin_acquire()
__attribute__ ((format(printf, 8, 9)))
static NV_STATUS __uvm_push_begin_acquire_with_info(uvm_channel_manager_t *manager, uvm_channel_type_t type, uvm_tracker_t *tracker,
        uvm_push_t *push,
        const char *filename, const char *function, int line, const char *format, ...)
{
    va_list args;

    NV_STATUS status = __uvm_push_begin_acquire(manager, type, tracker, push);
    if (status != NV_OK)
        return status;

    va_start(args, format);
    __uvm_push_fill_info(push, filename, function, line, format, args);
    va_end(args);

    return NV_OK;
}

// Internal helper for uvm_push_begin_on_channel()
__attribute__ ((format(printf, 6, 7)))
static NV_STATUS __uvm_push_begin_on_channel_with_info(uvm_channel_t *channel,
        uvm_push_t *push,
        const char *filename, const char *function, int line, const char *format, ...)
{
    va_list args;

    NV_STATUS status = __uvm_push_begin_on_channel(channel, push);
    if (status != NV_OK)
        return status;

    va_start(args, format);
    __uvm_push_fill_info(push, filename, function, line, format, args);
    va_end(args);

    return NV_OK;
}

// Begin a push on a channel of channel_type type
// Picks the first available channel. If all channels of the given type are
// busy, spin waits for one to become available.
//
// Notably channel_type can be UVM_CHANNEL_TYPE_ANY and that will pick any
// available channel.
//
// Notably requires a description of the push to be provided. This is currently
// unused, but will be in the future for tracking push history.
//
// Locking: on success acquires the concurrent push semaphore until uvm_push_end()
#define uvm_push_begin(manager, type, push, format, ...)                            \
    __uvm_push_begin_acquire_with_info((manager), (type), NULL, (push), \
        __FILE__, __FUNCTION__, __LINE__, (format), ##__VA_ARGS__)

// Begin a push on a channel of channel_type type with dependencies in the tracker
// This is equivalent to starting a push and acquiring the tracker, but in the
// future it will have the ability to pick the channel to do a push on in a
// smarter way based on its dependencies.
//
// Same as for uvm_push_acquire_tracker(), the tracker can be NULL. In this case
// this will be equivalent to just uvm_push_begin().
//
// Locking: on success acquires the concurrent push semaphore until uvm_push_end()
#define uvm_push_begin_acquire(manager, type, tracker, push, format, ...)       \
    __uvm_push_begin_acquire_with_info((manager), (type), (tracker), (push),    \
        __FILE__, __FUNCTION__, __LINE__, (format), ##__VA_ARGS__);

// Begin a push on a specific channel
// If the channel is busy, spin wait for it to become available.
//
// Locking: on success acquires the concurrent push semaphore until uvm_push_end()
#define uvm_push_begin_on_channel(channel, push, format, ...)                 \
    __uvm_push_begin_on_channel_with_info((channel), (push),                  \
        __FILE__, __FUNCTION__, __LINE__, (format), ##__VA_ARGS__);

// End a push
// Finishes the push and submits the methods to the GPU.
//
// This will always release the channel tracking semaphore with CE and that
// release can be affected by setting the push flags (commonly
// UVM_PUSH_FLAGS_CE_NEXT_FLUSH) prior to calling uvm_push_end().
//
// Notably doesn't wait for the push to complete on the GPU and is also
// guaranteed not to block waiting on any other GPU work to complete. The only
// contention that can happen is with other CPU threads updating channel and/or
// pushbuffer state, but all of these updates are expected to be fast.
//
// Completion of the push on the GPU can be tracked with a tracker by using
// uvm_tracker_add_push() or can be waited on directly with uvm_push_wait().
// Also see uvm_push_end_and_wait() that combines ending and waiting for a push.
//
// Locking: releases the concurrent push semaphore acquired in uvm_push_begin*()
void uvm_push_end(uvm_push_t *push);

// Wait for a push to complete its execution on the GPU.
//
// The push has to be finished prior to calling this function.
// Notably currently this will only check for errors on the channel the push has
// been made on while waiting for it to complete.
NV_STATUS uvm_push_wait(uvm_push_t *push);

// End a push and wait for it to complete execution on the GPU
// Shortcut for uvm_push_end() and uvm_push_wait().
NV_STATUS uvm_push_end_and_wait(uvm_push_t *push);

// Get the tracker entry tracking the push
// The push has to be finished before calling this function.
void uvm_push_get_tracker_entry(uvm_push_t *push, uvm_tracker_entry_t *entry);

// Acquire all the entries in the tracker.
// Subsequently pushed GPU work will not start before all the work tracked by
// tracker is complete.
// Notably a NULL tracker is handled the same way as an empty tracker.
void uvm_push_acquire_tracker(uvm_push_t *push, uvm_tracker_t *tracker);

// Acquire a single tracker entry
// Subsequently pushed GPU work will not start before the work tracked by
// tracker entry is complete.
void uvm_push_acquire_tracker_entry(uvm_push_t *push, uvm_tracker_entry_t *tracker_entry);

// Set a push flag
void uvm_push_set_flag(uvm_push_t *push, uvm_push_flag_t flag);

// Get and reset (if set) a push flag
bool uvm_push_get_and_reset_flag(uvm_push_t *push, uvm_push_flag_t flag);

// Get the size of the push so far
NvU32 uvm_push_get_size(uvm_push_t *push);

static uvm_gpu_t *uvm_push_get_gpu(uvm_push_t *push)
{
    UVM_ASSERT(push->channel);

    return uvm_channel_get_gpu(push->channel);
}

#endif // __UVM8_PUSH_H__
