/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_linux.h"
#include "uvm_common.h"
#include "uvm8_va_space.h"
#include "uvm8_va_range.h"
#include "uvm8_va_block.h"
#include "uvm8_mmu.h"
#include "uvm8_kvmalloc.h"

static struct kmem_cache *g_uvm_va_block_cache __read_mostly;

NV_STATUS uvm_va_block_init(void)
{
    g_uvm_va_block_cache = NV_KMEM_CACHE_CREATE("uvm_va_block_t", uvm_va_block_t);
    if (!g_uvm_va_block_cache)
        return NV_ERR_NO_MEMORY;
    return NV_OK;
}

void uvm_va_block_exit(void)
{
    kmem_cache_destroy_safe(&g_uvm_va_block_cache);
}

static size_t block_num_cpu_pages(uvm_va_block_t *block)
{
    return uvm_va_block_size(block) / PAGE_SIZE;
}

static size_t block_cpu_page_index(uvm_va_block_t *block, NvU64 addr)
{
    UVM_ASSERT(addr >= block->start);
    UVM_ASSERT(addr <= block->end);
    return (addr - block->start) / PAGE_SIZE;
}

NV_STATUS uvm_va_block_create(uvm_va_range_t *va_range,
                              NvU64 start,
                              NvU64 end,
                              uvm_va_block_t **out_block)
{
    uvm_va_block_t *block = NULL;
    NvU64 size = end - start + 1;
    NV_STATUS status;

    UVM_ASSERT(PAGE_ALIGNED(start));
    UVM_ASSERT(PAGE_ALIGNED(end + 1));
    UVM_ASSERT(PAGE_ALIGNED(size));
    UVM_ASSERT(size > 0);
    UVM_ASSERT(size <= UVM_VA_BLOCK_SIZE);
    UVM_ASSERT(start >= va_range->node.start);
    UVM_ASSERT(end <= va_range->node.end);

    // Blocks can't span a block alignment boundary
    UVM_ASSERT(UVM_VA_BLOCK_ALIGN_DOWN(start) == UVM_VA_BLOCK_ALIGN_DOWN(end));

    block = kmem_cache_zalloc(g_uvm_va_block_cache, NV_UVM_GFP_FLAGS);
    if (!block) {
        status = NV_ERR_NO_MEMORY;
        goto error;
    }

    kref_init(&block->kref);
    uvm_mutex_init(&block->lock, UVM_LOCK_ORDER_VA_BLOCK);
    block->start = start;
    block->end = end;
    block->va_range = va_range;

    if (block->va_range->type == UVM_VA_RANGE_TYPE_MMAP) {
        block->mmap.cpu.pages = uvm_kvmalloc_zero((size / PAGE_SIZE) * sizeof(block->mmap.cpu.pages[0]));
        if (!block->mmap.cpu.pages) {
            status = NV_ERR_NO_MEMORY;
            goto error;
        }
    }

    *out_block = block;
    return NV_OK;

error:
    uvm_va_block_release(block);
    return status;
}

// Looks up the current CPU mapping state of addr from the
// block->mmap.cpu.pte_bits bitmaps. If write access is enabled,
// UVM_PROT_READ_WRITE_ATOMIC is returned instead of UVM_PROT_READ_WRITE, since
// write access implies atomic access for CPUs.
static uvm_prot_t block_addr_prot_cpu(uvm_va_block_t *block, NvU64 addr)
{
    size_t page_index = block_cpu_page_index(block, addr);
    uvm_prot_t prot;

    UVM_ASSERT(block->va_range);
    UVM_ASSERT(block->va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    uvm_assert_mutex_locked(&block->lock);

    if (test_bit(page_index, block->mmap.cpu.pte_bits[UVM_PTE_BITS_CPU_WRITE]))
        prot = UVM_PROT_READ_WRITE_ATOMIC;
    else if (test_bit(page_index, block->mmap.cpu.pte_bits[UVM_PTE_BITS_CPU_READ]))
        prot = UVM_PROT_READ_ONLY;
    else
        prot = UVM_PROT_NONE;

    // Sanity checks
    if (prot != UVM_PROT_NONE) {
        UVM_ASSERT(uvm_processor_mask_test(&block->mmap.mapped, UVM_CPU_ID));

        // The CPU can't map the GPU directly, so the CPU must have a coherent
        // copy resident if mapped.
        UVM_ASSERT(uvm_processor_mask_test(&block->mmap.resident, UVM_CPU_ID));
        UVM_ASSERT(test_bit(page_index, block->mmap.cpu.resident));
        UVM_ASSERT(block->mmap.cpu.pages[page_index]);

        if (prot == UVM_PROT_READ_WRITE_ATOMIC) {
            // If write is set, read should be too
            UVM_ASSERT(test_bit(page_index, block->mmap.cpu.pte_bits[UVM_PTE_BITS_CPU_READ]));

            // If the prot is writable, then the CPU's page must be the only
            // coherent copy resident anywhere in the system.
            //
            // TODO: Assert that this page is not resident on any GPU.
        }
    }

    return prot;
}

// Returns true if the block has any valid CPU PTE mapping in the range given by
// [addr, addr + size).
//
// TODO: This might need to turn into a more comprehensive set of iterators,
//       such as for_each_valid_cpu_pte, for_each_writable_cpu_pte, etc.
static bool block_has_valid_mapping_cpu(uvm_va_block_t *block, NvU64 addr, size_t size)
{
    NvU64 valid_addr, outer_addr = addr + size;
    size_t addr_index, outer_index, valid_index;

    UVM_ASSERT(addr >= block->start);
    UVM_ASSERT(outer_addr - 1 <= block->end);

    // Early-out: check whether any address in this block has a CPU mapping
    if (!uvm_processor_mask_test(&block->mmap.mapped, UVM_CPU_ID)) {
        UVM_ASSERT(bitmap_empty(block->mmap.cpu.pte_bits[UVM_PTE_BITS_CPU_READ], PAGES_PER_UVM_VA_BLOCK));
        UVM_ASSERT(bitmap_empty(block->mmap.cpu.pte_bits[UVM_PTE_BITS_CPU_WRITE], PAGES_PER_UVM_VA_BLOCK));
        return false;
    }

    addr_index = block_cpu_page_index(block, addr);

    // block_cpu_page_index will assert when outer_addr exceeds block->end
    outer_index = (outer_addr - block->start) / PAGE_SIZE;

    // All valid mappings have at least read permissions so we only need to
    // inspect the read bits.
    valid_index = find_next_bit(block->mmap.cpu.pte_bits[UVM_PTE_BITS_CPU_READ],
                                outer_index,
                                addr_index);

    if (valid_index == outer_index)
        return false;

    valid_addr = block->start + valid_index * PAGE_SIZE;
    UVM_ASSERT(block_addr_prot_cpu(block, valid_addr) != UVM_PROT_NONE);
    return true;
}

// Unmap CPU PTEs without freeing any pages. This is safe to call even if we
// don't hold a reference on the mm owning the vma.
static void block_unmap_cpu(uvm_va_block_t *block, NvU64 start_addr, size_t size)
{
    uvm_va_range_t *va_range = block->va_range;
    NvU64 outer_addr = start_addr + size;
    size_t start_index, num_pages;
    uvm_pte_bits_cpu_t pte_bit;

    UVM_ASSERT(va_range);
    UVM_ASSERT(va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    UVM_ASSERT(PAGE_ALIGNED(start_addr));
    UVM_ASSERT(PAGE_ALIGNED(size));
    UVM_ASSERT(start_addr >= block->start);
    UVM_ASSERT(outer_addr - 1 <= block->end);
    uvm_assert_mutex_locked(&block->lock);

    if (!va_range->va_space)
        return;

    start_index = block_cpu_page_index(block, start_addr);

    // Early-out if nothing in the range is mapped
    if (!block_has_valid_mapping_cpu(block, start_addr, size))
        return;

    unmap_mapping_range(&va_range->va_space->mapping, start_addr, size, 1);

    // Clear block PTE state
    num_pages = size / PAGE_SIZE;
    for (pte_bit = 0; pte_bit < UVM_PTE_BITS_CPU_MAX; pte_bit++)
        bitmap_clear(block->mmap.cpu.pte_bits[pte_bit], start_index, num_pages);

    // Check whether the block has any more mappings
    if (bitmap_empty(block->mmap.cpu.pte_bits[UVM_PTE_BITS_CPU_READ], PAGES_PER_UVM_VA_BLOCK)) {
        UVM_ASSERT(bitmap_empty(block->mmap.cpu.pte_bits[UVM_PTE_BITS_CPU_WRITE], PAGES_PER_UVM_VA_BLOCK));
        uvm_processor_mask_clear(&block->mmap.mapped, UVM_CPU_ID);
    }
}

// This function essentially works as a wrapper around vm_insert_page (hence
// the similar function prototype). This is needed since vm_insert_page
// doesn't take permissions as input, but uses vma->vm_page_prot instead.
// Since we may have multiple VA blocks under one VMA which need to map
// with different permissions, we have to manually change vma->vm_page_prot for
// each call to vm_insert_page. Multiple fauts under one VMA in separate
// blocks can be serviced concurrently, so the VMA wrapper lock is used
// to protect access to vma->vm_page_prot.
static NV_STATUS uvm_cpu_insert_page(struct vm_area_struct *vma,
                                     NvU64 addr,
                                     struct page *page,
                                     uvm_prot_t new_prot)
{
    uvm_vma_wrapper_t *vma_wrapper;
    unsigned long target_flags;
    pgprot_t target_pgprot;
    int ret;

    UVM_ASSERT(vma);
    UVM_ASSERT(vma->vm_private_data);

    vma_wrapper = vma->vm_private_data;
    target_flags = vma->vm_flags;

    if (new_prot == UVM_PROT_READ_ONLY)
        target_flags &= ~VM_WRITE;

    target_pgprot = vm_get_page_prot(target_flags);

    // Take VMA wrapper lock to check vma->vm_page_prot
    uvm_down_read(&vma_wrapper->lock);

    // Take a write lock if we need to modify the VMA vm_page_prot
    // - vma->vm_page_prot creates writable PTEs but new prot is RO
    // - vma->vm_page_prot creates read-only PTEs but new_prot is RW
    if (pgprot_val(vma->vm_page_prot) != pgprot_val(target_pgprot)) {
        uvm_up_read(&vma_wrapper->lock);
        uvm_down_write(&vma_wrapper->lock);

        vma->vm_page_prot = target_pgprot;

        uvm_downgrade_write(&vma_wrapper->lock);
    }

    ret = vm_insert_page(vma, addr, page);
    uvm_up_read(&vma_wrapper->lock);
    if (ret) {
        UVM_ASSERT_MSG(ret == -ENOMEM, "ret: %d\n", ret);
        return errno_to_nv_status(ret);
    }

    return NV_OK;
}

// Creates or modifies a CPU mapping for the given virtual address and updates
// the block's mapping and pte_bits bitmaps to reflect the new mapping.
//
// new_prot must not be UVM_PROT_NONE. block_unmap_cpu should be used instead.
//
// If the existing mapping already is new_prot, the behavior depends on
// force_remap. If force_remap is false this is a no-op. If force_remap is true
// the existing CPU mapping is unmapped first and the identical mapping is re-
// inserted. This handles cases in which the block tracking goes stale because
// the kernel unmapped the CPU behind our back.
//
// It is the caller's responsibility to:
//  - Guarantee that vm_insert_page is safe to use (vma->vm_mm has a reference
//    and mmap_sem is held in at least read mode)
//  - Ensure that the CPU page exists in block->mmap.cpu.pages
//  - Manage the block's residency bitmap
//  - Decide whether force_remap is appropriate
//  - Ensure that the block hasn't been killed (block->va_range is present)
static NV_STATUS block_map_cpu_page(uvm_va_block_t *block,
                                    NvU64 addr,
                                    uvm_prot_t new_prot,
                                    bool force_remap)
{
    uvm_va_range_t *va_range = block->va_range;
    uvm_prot_t curr_prot;
    size_t page_index;
    struct vm_area_struct *vma;
    NV_STATUS status;

    UVM_ASSERT(va_range);
    UVM_ASSERT(va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    UVM_ASSERT(addr >= block->start);
    UVM_ASSERT(addr <= block->end);
    UVM_ASSERT(PAGE_ALIGNED(addr));
    UVM_ASSERT(new_prot != UVM_PROT_NONE);
    UVM_ASSERT(new_prot < UVM_PROT_MAX);

    // It's possible here that current->mm != vma->vm_mm. That can happen for
    // example due to access_process_vm (ptrace) or get_user_pages from another
    // driver. In those cases the caller has taken care of ref counting
    // vma->vm_mm for us, so we can safely operate on the vma, but we can't use
    // uvm_va_range_vma_current.
    vma = uvm_va_range_vma(va_range);
    uvm_assert_mmap_sem_locked_read(&vma->vm_mm->mmap_sem);

    uvm_assert_mutex_locked(&block->lock);

    page_index = block_cpu_page_index(block, addr);
    curr_prot = block_addr_prot_cpu(block, addr);

    // For the CPU, write implies atomic
    if (new_prot == UVM_PROT_READ_WRITE)
        new_prot = UVM_PROT_READ_WRITE_ATOMIC;

    // Check for existing VMA permissions
    // They can be modified after the initial mmap by mprotect
    if (!(vma->vm_flags & VM_READ))
        return NV_ERR_INVALID_ACCESS_TYPE;
    if (new_prot == UVM_PROT_READ_WRITE_ATOMIC && !(vma->vm_flags & VM_WRITE))
        return NV_ERR_INVALID_ACCESS_TYPE;

    if (!force_remap && curr_prot == new_prot)
        return NV_OK;

    // This unmap handles both the force_remap case and a special case required
    // by vm_insert_page. vm_insert_page returns -EBUSY when there's already a
    // mapping present at fault_addr, so we have to unmap first anyway when
    // upgrading from RO -> RW or downgrading from RW -> RO.
    if (curr_prot != UVM_PROT_NONE)
        unmap_mapping_range(&va_range->va_space->mapping, addr, PAGE_SIZE, 1);

    status = uvm_cpu_insert_page(vma, addr, block->mmap.cpu.pages[page_index], new_prot);
    if (status != NV_OK)
        return status;

    // Update block mapping state
    if (curr_prot != new_prot) {
        // Transitioning from Invalid -> RO or Invalid -> RW
        if (curr_prot == UVM_PROT_NONE) {
            uvm_processor_mask_set(&block->mmap.mapped, UVM_CPU_ID);
            __set_bit(page_index, block->mmap.cpu.pte_bits[UVM_PTE_BITS_CPU_READ]);
        }

        // Transitioning from Invalid -> RW or RO -> RW
        if (new_prot == UVM_PROT_READ_WRITE_ATOMIC)
            __set_bit(page_index, block->mmap.cpu.pte_bits[UVM_PTE_BITS_CPU_WRITE]);

        // On RW -> RO transition, clear UVM_PTE_BITS_CPU_WRITE bit
        if (curr_prot == UVM_PROT_READ_WRITE_ATOMIC && new_prot == UVM_PROT_READ_ONLY)
            __clear_bit(page_index, block->mmap.cpu.pte_bits[UVM_PTE_BITS_CPU_WRITE]);
    }

    return NV_OK;
}

// TODO: Implement block_map_cpu_range for optimizing population and migration
//       of already-mapped CPU regions.

// Tears down everything within the block, but doesn't free the block itself.
// Note that when uvm_va_block_kill is called, this is called twice: once for
// the initial kill itself, then again when the block's ref count is eventually
// destroyed. block->va_range is used to track whether the block has already
// been killed.
static void block_kill(uvm_va_block_t *block)
{
    uvm_va_range_t *va_range = block->va_range;
    size_t i;

    if (!va_range)
        return;

    if (va_range->type == UVM_VA_RANGE_TYPE_MMAP) {
        NV_STATUS status;
        uvm_perf_event_data_t event_data;
        event_data.block_create.block = block;

        status = uvm_perf_event_notify(&va_range->va_space->perf_events, UVM_PERF_EVENT_BLOCK_DESTROY, &event_data);
        UVM_ASSERT(status == NV_OK);

        block_unmap_cpu(block, block->start, uvm_va_block_size(block));

        // Free CPU pages
        if (block->mmap.cpu.pages) {
            for (i = 0; i < block_num_cpu_pages(block); i++) {
                if (block->mmap.cpu.pages[i])
                    __free_page(block->mmap.cpu.pages[i]);
                else
                    UVM_ASSERT(!test_bit(i, block->mmap.cpu.resident));
            }
            uvm_kvfree(block->mmap.cpu.pages);
        }

        // TODO: Unmap and free GPU pages
    }

    block->va_range = NULL;
}

// Called when the block's ref count drops to 0
void uvm_va_block_destroy(struct kref *kref)
{
    uvm_va_block_t *block = container_of(kref, uvm_va_block_t, kref);

    // Nobody else should have a reference when freeing
    uvm_assert_mutex_unlocked(&block->lock);

    uvm_mutex_lock(&block->lock);
    block_kill(block);
    uvm_mutex_unlock(&block->lock);
    kmem_cache_free(g_uvm_va_block_cache, block);
}

void uvm_va_block_kill(uvm_va_block_t *va_block)
{
    uvm_mutex_lock(&va_block->lock);
    block_kill(va_block);
    uvm_mutex_unlock(&va_block->lock);

    // May call block_kill again
    uvm_va_block_release(va_block);
}

// Re-calculate the block's top-level processor masks:
//   - block->mmap.mapped
//   - block->mmap.resident
//
// This is called on block split.
static void block_set_processor_masks(uvm_va_block_t *block)
{
    size_t num_pages = block_num_cpu_pages(block);

    if (bitmap_empty(block->mmap.cpu.pte_bits[UVM_PTE_BITS_CPU_READ], num_pages)) {
        UVM_ASSERT(bitmap_empty(block->mmap.cpu.pte_bits[UVM_PTE_BITS_CPU_WRITE], num_pages));
        uvm_processor_mask_clear(&block->mmap.mapped, UVM_CPU_ID);
    }
    else {
        uvm_processor_mask_set(&block->mmap.mapped, UVM_CPU_ID);
    }

    if (bitmap_empty(block->mmap.cpu.resident, num_pages)) {
        UVM_ASSERT(!uvm_processor_mask_test(&block->mmap.mapped, UVM_CPU_ID));
        uvm_processor_mask_clear(&block->mmap.resident, UVM_CPU_ID);
    }
    else {
        uvm_processor_mask_set(&block->mmap.resident, UVM_CPU_ID);
    }
}

// Split the CPU state within the existing block. existing's start is correct
// but its end has not yet been adjusted.
static void uvm_va_block_split_cpu(uvm_va_block_t *existing, uvm_va_block_t *new)
{
    struct page **temp_pages;
    size_t existing_pages, new_pages = block_num_cpu_pages(new);
    uvm_pte_bits_cpu_t pte_bit;

    UVM_ASSERT(existing->va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    UVM_ASSERT(existing->va_range->type == new->va_range->type);
    UVM_ASSERT(existing->start < new->start);
    UVM_ASSERT(existing->end == new->end);
    UVM_ASSERT(existing->mmap.cpu.pages);
    UVM_ASSERT(new->mmap.cpu.pages);

    // We don't have to unmap the CPU since its virtual -> physical mappings
    // don't change.

    UVM_ASSERT(PAGE_ALIGNED(new->start));
    UVM_ASSERT(PAGE_ALIGNED(existing->start));
    existing_pages = (new->start - existing->start) / PAGE_SIZE;

    // Move pages from the top of existing down into new
    memcpy(&new->mmap.cpu.pages[0],
           &existing->mmap.cpu.pages[existing_pages],
           block_num_cpu_pages(new) * sizeof(new->mmap.cpu.pages[0]));

    // Attempt to shrink existing's pages allocation. If the realloc fails, just
    // keep on using the old larger one.
    temp_pages = uvm_kvrealloc(existing->mmap.cpu.pages,
                               existing_pages * sizeof(existing->mmap.cpu.pages[0]));
    if (temp_pages)
        existing->mmap.cpu.pages = temp_pages;

    // The new block is always in the upper region of existing, so shift the bit
    // vectors down.
    //
    // Note that bitmap_shift_right requires both dst and src to be the same
    // size. That's ok since we don't scale them by block size.
    bitmap_shift_right(new->mmap.cpu.resident,
                       existing->mmap.cpu.resident,
                       existing_pages,
                       PAGES_PER_UVM_VA_BLOCK);
    bitmap_clear(existing->mmap.cpu.resident, existing_pages, new_pages);

    for (pte_bit = 0; pte_bit < UVM_PTE_BITS_CPU_MAX; pte_bit++) {
        bitmap_shift_right(new->mmap.cpu.pte_bits[pte_bit],
                           existing->mmap.cpu.pte_bits[pte_bit],
                           existing_pages,
                           PAGES_PER_UVM_VA_BLOCK);
        bitmap_clear(existing->mmap.cpu.pte_bits[pte_bit], existing_pages, new_pages);
    }

    // block_set_processor_masks calculates the new masks based on the size of
    // the block, which in the case of existing has not been updated yet. That's
    // ok because we cleared out the now-excess bits above.
    block_set_processor_masks(existing);
    block_set_processor_masks(new);
}

NV_STATUS uvm_va_block_split(uvm_va_block_t *existing_va_block,
                             NvU64 new_end,
                             uvm_va_block_t **new_va_block,
                             uvm_va_range_t *new_va_range)
{
    uvm_va_space_t *va_space;
    uvm_va_block_t *new_block = NULL;
    NV_STATUS status;

    // Be careful, new_va_range's va_space isn't set yet
    UVM_ASSERT(existing_va_block->va_range);
    va_space = existing_va_block->va_range->va_space;

    // External range types can't be split
    UVM_ASSERT(existing_va_block->va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    UVM_ASSERT(new_va_range->type == UVM_VA_RANGE_TYPE_MMAP);
    uvm_assert_rwsem_locked_write(&va_space->lock);

    UVM_ASSERT(new_end > existing_va_block->start);
    UVM_ASSERT(new_end < existing_va_block->end);
    UVM_ASSERT(PAGE_ALIGNED(new_end + 1));

    status = uvm_va_block_create(new_va_range, new_end + 1, existing_va_block->end, &new_block);
    if (status != NV_OK)
        return status;

    // We're protected from other splits and faults by the va_space lock being
    // held in write mode, but that doesn't stop the reverse mapping (eviction
    // path) from inspecting the existing block. Stop them by taking the block
    // lock. When a reverse mapping thread takes this lock afterwards, it will
    // have to re-inspect state and may see that it should use the newly-split
    // block instead.
    uvm_mutex_lock(&existing_va_block->lock);

    // As soon as we update existing's reverse mappings to point to the newly-
    // split block, the eviction path could try to operate on the new block.
    // Lock that out too until new is ready.
    //
    // Note that we usually shouldn't nest block locks, but it's ok here because
    // we just created new_block so no other thead could possibly take it out of
    // order with existing's lock.
    uvm_mutex_lock_no_tracking(&new_block->lock);

    // TODO: This modifies existing. That's ok for now because there are no
    //       failure points after this, but there may be in the future when we
    //       have to split GPU state. If that happens, this will have to be made
    //       transactional.
    uvm_va_block_split_cpu(existing_va_block, new_block);

    // TODO: Split and move GPU page tables and pages within block

    // Update existing's state only once we're past all failure points
    existing_va_block->end = new_end;

    uvm_mutex_unlock_no_tracking(&new_block->lock);
    uvm_mutex_unlock(&existing_va_block->lock);

    if (new_va_block)
        *new_va_block = new_block;
    return NV_OK;
}

NV_STATUS uvm_va_block_cpu_fault(uvm_va_block_t *va_block, NvU64 fault_addr, bool is_write)
{
    uvm_va_range_t *va_range = va_block->va_range;
    size_t page_index;
    struct page *page;
    NV_STATUS status = NV_OK;
    bool force_remap;
    uvm_prot_t page_prot = UVM_PROT_READ_WRITE_ATOMIC;

    UVM_ASSERT(va_range);
    uvm_assert_rwsem_locked_read(&va_range->va_space->lock);
    UVM_ASSERT(va_range->type == UVM_VA_RANGE_TYPE_MMAP);

    UVM_ASSERT(fault_addr >= va_block->start);
    UVM_ASSERT(fault_addr <= va_block->end);

    // CPU faults to non-migratable range groups are always fatal, even if the
    // page is already on the CPU.
    if (va_range->range_group && !atomic_read(&va_range->range_group->allow_migration))
        return NV_ERR_INVALID_OPERATION;

    // If the VA range is marked with read duplication enabled and this is a read
    // access, map the range as read-only.
    // TODO: this should be moved to a heuristics function eventually
    if (va_range->read_duplication == UVM_READ_DUPLICATION_ENABLED && !is_write)
        page_prot = UVM_PROT_READ_ONLY;

    // We have to use vm_insert_page instead of handing the page to the kernel
    // and letting it insert the mapping, and we must do that while holding the
    // lock on this VA block. Otherwise there will be a window in which we think
    // we've mapped the page but the CPU mapping hasn't actually been created
    // yet. During that window a GPU fault event could arrive and claim
    // ownership of that VA, "unmapping" it. Then later the kernel would
    // eventually establish the mapping, and we'd end up with both CPU and GPU
    // thinking they each owned the page.
    //
    // This function must only be called when it's safe to call vm_insert_page.
    // That is, there's a reference held on the vma's vm_mm and vm_mm->mmap_sem
    // is held in at least read mode, but current->mm might not be vma->vm_mm.

    page_index = block_cpu_page_index(va_block, fault_addr);

    uvm_mutex_lock(&va_block->lock);

    // TODO: Move the below logic to some sort of
    //       block_make_cpu_resident(.reason = CPU_FAULT) API that queries
    //       policy state for an action, then performs any migrations and
    //       mappings necessary for that action.

    page = va_block->mmap.cpu.pages[page_index];
    if (!page) {
        page = alloc_pages(NV_UVM_GFP_FLAGS | GFP_HIGHUSER | __GFP_ZERO, 0);
        if (!page) {
            status = NV_ERR_NO_MEMORY;
            goto out;
        }

        // Update residency state
        va_block->mmap.cpu.pages[page_index] = page;
        uvm_processor_mask_set(&va_block->mmap.resident, UVM_CPU_ID);
        __set_bit(page_index, va_block->mmap.cpu.resident);

    }
    else {
        UVM_ASSERT(uvm_processor_mask_test(&va_block->mmap.resident, UVM_CPU_ID));
        UVM_ASSERT(test_bit(page_index, va_block->mmap.cpu.resident));
    }


    // TODO: The kernel can downgrade our CPU mappings at any time without
    //       notifying us, which means our PTE state could be stale. We handle
    //       this for now by always forcibly re-mapping the CPU PTE, but we
    //       could instead track if we get a bunch of unexpected faults, and
    //       only forcibly unmap once we've exceeded some threshold.
    //
    //       A CPU fault is unexpected if:
    //       curr_prot == RW || (!is_write && curr_prot == RO)
    //
    //       But, note that we could also get "unexpected" faults if multiple
    //       CPU threads fault on the same address simultaneously and race to
    //       create the mapping.
    force_remap = true;

    status = block_map_cpu_page(va_block, fault_addr, page_prot, force_remap);
    if (status != NV_OK) {
        // No need to free page. It'll get freed on teardown of this block
        goto out;
    }

out:
    uvm_mutex_unlock(&va_block->lock);
    return status;
}

NV_STATUS uvm_va_block_find(uvm_va_space_t *va_space, NvU64 addr, uvm_va_block_t **out_block)
{
    uvm_va_range_t *va_range;
    uvm_va_block_t *block;
    size_t index;

    va_range = uvm_va_range_find(va_space, addr);
    if (!va_range)
        return NV_ERR_INVALID_ADDRESS;

    index = uvm_va_range_block_index(va_range, addr);
    block = uvm_va_range_block(va_range, index);
    if (!block)
        return NV_ERR_OBJECT_NOT_FOUND;

    *out_block = block;
    return NV_OK;
}

NV_STATUS uvm_va_block_find_create(uvm_va_space_t *va_space, NvU64 addr, uvm_va_block_t **out_block)
{
    uvm_va_range_t *va_range;
    NvU64 index;

    va_range = uvm_va_range_find(va_space, addr);
    if (!va_range)
        return NV_ERR_INVALID_ADDRESS;

    index = uvm_va_range_block_index(va_range, addr);
    return uvm_va_range_block_create(va_range, index, out_block);
}

NV_STATUS uvm8_test_change_pte_mapping(UVM_TEST_CHANGE_PTE_MAPPING_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    uvm_va_block_t *block;
    NV_STATUS status = NV_OK;
    size_t page_index;
    uvm_prot_t new_prot;

    // TODO: For now we only handle CPU mappings
    if (!uvm_uuid_is_cpu(&params->uuid)) {
        UVM_ERR_PRINT("GPU mappings are not currently supported\n");
        return NV_ERR_NOT_SUPPORTED;
    }

    if (!PAGE_ALIGNED(params->va))
        return NV_ERR_INVALID_ADDRESS;

    // mmap_sem isn't needed for invalidating CPU mappings, but it will be
    // needed for inserting them.
    uvm_down_read_mmap_sem(&current->mm->mmap_sem);
    uvm_down_read(&va_space->lock);

    status = uvm_va_block_find_create(va_space, params->va, &block);
    if (status != NV_OK)
        goto out;

    if (block->va_range->type != UVM_VA_RANGE_TYPE_MMAP) {
        status = NV_ERR_INVALID_ADDRESS;
        goto out;
    }

    uvm_mutex_lock(&block->lock);
    switch (params->mapping) {
        case UVM_TEST_PTE_MAPPING_INVALID:
            block_unmap_cpu(block, params->va, PAGE_SIZE);
            break;

        case UVM_TEST_PTE_MAPPING_READ_ONLY:
        case UVM_TEST_PTE_MAPPING_READ_WRITE:
        case UVM_TEST_PTE_MAPPING_READ_WRITE_ATOMIC:
            new_prot = (params->mapping == UVM_TEST_PTE_MAPPING_READ_ONLY) ?
                UVM_PROT_READ_ONLY : UVM_PROT_READ_WRITE_ATOMIC;

            // Creating new CPU mappings requires that we hold a reference on
            // vma->vm_mm, but we can only do that if current->mm == vma->vm_mm.
            if (!uvm_va_range_vma_current(block->va_range)) {
                status = NV_ERR_INVALID_STATE;
                break;
            }

            // The page must be resident on the CPU to map
            page_index = block_cpu_page_index(block, params->va);
            if (!test_bit(page_index, block->mmap.cpu.resident)) {
                status = NV_ERR_INVALID_OPERATION;
                break;
            }

            status = block_map_cpu_page(block, params->va, new_prot, true);
            break;
        default:
            status = NV_ERR_INVALID_ARGUMENT;
            break;
    }
    uvm_mutex_unlock(&block->lock);

out:
    uvm_up_read(&va_space->lock);
    uvm_up_read_mmap_sem(&current->mm->mmap_sem);
    return status;
}

NV_STATUS uvm8_test_va_block_info(UVM_TEST_VA_BLOCK_INFO_PARAMS *params, struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    uvm_va_block_t *va_block;
    NV_STATUS status = NV_OK;

    BUILD_BUG_ON(UVM_TEST_VA_BLOCK_SIZE != UVM_VA_BLOCK_SIZE);

    uvm_down_read_mmap_sem(&current->mm->mmap_sem);
    uvm_down_read(&va_space->lock);

    status = uvm_va_block_find(va_space, params->lookup_address, &va_block);
    if (status != NV_OK)
        goto out;

    if (!uvm_va_range_vma_current(va_block->va_range)) {
        // We aren't in the same mm as the one which owns the vma
        status = NV_ERR_INVALID_STATE;
        goto out;
    }

    params->va_block_start = va_block->start;
    params->va_block_end   = va_block->end;

out:
    uvm_up_read(&va_space->lock);
    uvm_up_read_mmap_sem(&current->mm->mmap_sem);
    return status;
}
