/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#ifndef __UVM8_VA_BLOCK_H__
#define __UVM8_VA_BLOCK_H__

#include "uvm8_forward_decl.h"
#include "uvmtypes.h"
#include "uvm_linux.h"
#include "uvm_perf.h"
#include "uvm8_gpu.h"
#include "uvm8_lock.h"
#include "uvm8_test_ioctl.h"

// VA blocks are the leaf nodes in the uvm_va_space tree:
//
//  uvm_va_space -> uvm_va_range -> uvm_va_block
//
// Each VA block is contained within a single VA range, and contains state on
// VAs covered by that block. Most importantly, the block tracks the current
// state of the virtual-to-physical mappings for all VAs within that block
// across all processors, along with the physical residency location for each
// VA.
//
// The block serializes both CPU and GPU operations on all VAs under that block.
// The CPU work is serialized with the block lock, and the GPU work is
// serialized by the block work tracker which itself is protected by the block
// lock.
//
// The size of each block varies from the size of the smallest VA range
// (PAGE_SIZE) to the max block size specified by UVM_VA_BLOCK_BITS. No block
// will span a 2^UVM_VA_BLOCK_BITS boundary in VA space. The size of the block
// is determined by the alignment of the parent VA range and the block's
// placement within the range.
//
// Note that this means user space will get best allocation efficiency if it
// allocates memory in 2^UVM_VA_BLOCK_BITS naturally-aligned chunks.
//
// UVM_VA_BLOCK_BITS is 21, meaning the maximum block size is 2MB. Rationale:
// - 2MB matches the largest Pascal GPU page size so it's a natural fit
// - 2MB won't span more than one PDE on any chip, so the VA blocks never need
//   to track more than a single GPU PDE.
// - 2MB is a decent tradeoff between memory overhead and serialization
//   contention.
//
#define UVM_VA_BLOCK_BITS               21

// Max size of a block in bytes
#define UVM_VA_BLOCK_SIZE               ((NvLength)1 << UVM_VA_BLOCK_BITS)

// Mask to clear the lower bits
#define UVM_VA_BLOCK_MASK               (~(UVM_VA_BLOCK_SIZE - 1))

#define UVM_VA_BLOCK_ALIGN_DOWN(addr)   ((addr) & UVM_VA_BLOCK_MASK)
#define UVM_VA_BLOCK_ALIGN_UP(addr)     (((addr) + UVM_VA_BLOCK_SIZE - 1) & UVM_VA_BLOCK_MASK)

#define PAGES_PER_UVM_VA_BLOCK          (UVM_VA_BLOCK_SIZE / PAGE_SIZE)

// enums used for indexing into the array of pte_bits bitmaps in the VA block
// which hold the current state of each PTE. For a given {processor, PTE}, the
// bits represented here must be enough to re-create the non-address portion of
// the PTE for that processor.

// If _READ is not set, the PTE mapping is not valid.
// If _WRITE is set, _READ is also set (_WRITE implies _READ).
typedef enum
{
    UVM_PTE_BITS_CPU_READ,
    UVM_PTE_BITS_CPU_WRITE,
    UVM_PTE_BITS_CPU_MAX
} uvm_pte_bits_cpu_t;

// TODO: uvm_pte_bits_gpu_t will also include ATOMIC and VOLATILE values

typedef struct
{
    // Per-processor residency bit vector, used for fast lookup of which
    // processors are active in this block.
    //
    // A set bit means the corresponding processor has a coherent physical copy
    // of memory somewhere in the block. The per-processor state must then be
    // inspected to find out which page. The processor may or may not have a
    // mapping to that physical memory, however.
    //
    // A cleared bit means the corresponding processor does not have a coherent
    // physical copy of any pages under this block. The processor may still have
    // cached pages allocated for future use, however. It also may have mappings
    // to pages resident on other processors.
    uvm_processor_mask_t resident;

    // Per-processor mapping bit vector, used for fast lookup of which
    // processors are active in this block.
    //
    // A set bit means the corresponding processor has an active, valid page
    // table mapping to some VA in this block. The per-processor pte_bits state
    // must then be inspected to find out the mapping address and permissions.
    //
    // A cleared bit means the corresponding processor has no virtual mappings
    // within this block (all pte_bits entries are 0).
    uvm_processor_mask_t mapped;

    struct
    {
        // Per-page residency bit vector, used for fast traversal of resident
        // pages.
        //
        // A set bit means the CPU has a coherent copy of the physical page
        // resident in its memory, and that the corresponding entry in the pages
        // array is present. This does not mean that the coherent copy is
        // currently mapped anywhere, however. A page may be resident on
        // multiple processors when in read-duplicate mode.
        //
        // A cleared bit means the CPU does not have a coherent copy of that
        // page resident. The corresponding entry in the pages array may or may
        // not present. If the entry is present, it's a cached page which can be
        // reused in the future.
        //
        // Allocating PAGES_PER_UVM_VA_BLOCK is overkill when the block is
        // smaller than UVM_VA_BLOCK_SIZE, but it's not much extra memory
        // overhead on the whole.
        DECLARE_BITMAP(resident, PAGES_PER_UVM_VA_BLOCK);

        // Per-page array of physical pages. This array scales dynamically with
        // the block size.
        //
        // TODO: These should be some UVM CPU page data structure, complete with
        //       per-GPU IOMMU mappings when applicable. We shouldn't track CPU
        //       PFNs for the same reason.
        struct page **pages;

        // Per-page mapping bit vectors, one per bit we need to track. These are
        // used for fast traversal of valid mappings in the block. These contain
        // all non-address bits needed to establish a virtual mapping on this
        // processor (permissions, cacheability, etc).
        //
        // A cleared bit in UVM_PTE_BITS_CPU_READ means the CPU has no valid
        // virtual mapping to that address (the access will fault). Further,
        // UVM_PTE_BITS_CPU_WRITE is guaranteed to also be clear.
        //
        // A set bit in UVM_PTE_BITS_CPU_READ means the CPU has a valid mapping
        // at that address with at least read permissions. The physical page for
        // that mapping is contained in the pages array. If
        // UVM_PTE_BITS_CPU_WRITE is not set, the mapping is read-only.
        // Otherwise, the mapping is read-write.
        //
        // Note that this is the maximum permissions a PTE could have, but not
        // necessarily the actual current permissions of the CPU PTEs. The UVM
        // driver will never change the PTEs without updating this state, but
        // the kernel can downgrade our CPU mappings at any time without
        // notifying the UVM driver (for example in response to user space
        // calling madvise with MADV_DONTNEED).
        //
        // We can't use DECLARE_BITMAP since it doesn't allow arrays of bitmaps.
        unsigned long pte_bits[UVM_PTE_BITS_CPU_MAX][BITS_TO_LONGS(PAGES_PER_UVM_VA_BLOCK)];
    } cpu;

    // TODO: Per-GPU residency and mapping state

    uvm_perf_modules_data_t perf_modules_data;
} uvm_va_block_mmap_t;

// TODO: Worst-case we could have one of these per system page. Options:
//       1) Rely on the OOM killer to prevent the user from trying to do that
//       2) Be much more space-conscious in this struct (difficult)
//       3) Cap the per-process range and/or block count, like vm.max_map_count
//          does for vmas
struct uvm_va_block_struct
{
    // Reference count for this block. References are held by:
    // - The parent VA range
    // - The reverse map
    // - The eviction path temporarily when attempting to evict a GPU page under
    //   this block
    //
    // This isn't protected by the lock on the eviction path, so it must be
    // atomic. kref provides that.
    struct kref kref;

    // Lock protecting the block. See the comment at the top of uvm8.c.
    uvm_mutex_t lock;

    // Parent VA range. If this is not set, the block is considered dead and is
    // waiting for the last ref count to be removed.
    //
    // TODO: Since the va_range contains the policy for this block (preferred
    //       location, accessed-by, etc) the parent va_range must be inspected
    //       for operations on this block. That's ok for migrations and GPU
    //       faults since they lock the va_space first, but it's not ok for the
    //       eviction path which only locks the block. We'll have to figure out
    //       how to handle that, perhaps by having block operations take a
    //       snapshot of the range's policy state.
    //
    // This field can be read while holding either the block lock or just the VA
    // space lock in read mode, since it can only change when the VA space lock
    // is held in write mode.
    uvm_va_range_t *va_range;

    // Virtual address [start, end] covered by this block. These fields can be
    // read while holding either the block lock or just the VA space lock in
    // read mode, since they can only change when the VA space lock is held in
    // write mode.
    NvU64 start;
    NvU64 end;

    // The type to use for the union is given by va_range->type
    union
    {
        uvm_va_block_mmap_t mmap;
        // More here eventually
    };
};

// Module load/exit
NV_STATUS uvm_va_block_init(void);
void uvm_va_block_exit(void);

// Allocates and initializes the block. The block's ref count is initialized to
// 1. The caller is responsible for inserting the block into its parent
// va_range.
//
// The caller must be holding the VA space lock in at least read mode.
NV_STATUS uvm_va_block_create(uvm_va_range_t *va_range,
                              NvU64 start,
                              NvU64 end,
                              uvm_va_block_t **out_block);

// Internal function called only when uvm_va_block_release drops the ref count
// to 0. Do not call directly.
void uvm_va_block_destroy(struct kref *kref);

static inline void uvm_va_block_retain(uvm_va_block_t *va_block)
{
    kref_get(&va_block->kref);
}

static inline void uvm_va_block_release(uvm_va_block_t *va_block)
{
    if (va_block) {
        uvm_assert_mutex_unlocked(&va_block->lock);
        kref_put(&va_block->kref, uvm_va_block_destroy);
    }
}

// Unmaps all memory associated with the block and drops the ref count of the
// block. This allows the caller to free resources associated with this block
// regardless of the block's current ref count. Most importantly it allows the
// VA covered by this block to be immediately available for other page table
// mappings upon return.
//
// This clears block->va_range, so only the VA range destroy path should call
// it. Other paths with references on this block, specifically the eviction path
// which temporarily takes a reference to the block, must always check the block
// state after taking the block lock to see if their mapping is still in place.
//
// All of the unmap and state destruction steps are also performed when the ref
// count goes to 0, so this function only needs to be called if the block's
// resources need to be reclaimed immediately.
//
// The caller should not lock the block before calling this function.
//
// This performs a uvm_va_block_release.
void uvm_va_block_kill(uvm_va_block_t *va_block);

// Exactly the same split semantics as uvm_va_range_split, including error
// handling. See that function's comments for details.
//
// new_va_block's va_range is set to new_va_range before any reverse mapping is
// established to the new block, but the caller is responsible for inserting the
// new block into the range.
NV_STATUS uvm_va_block_split(uvm_va_block_t *existing_va_block,
                             NvU64 new_end,
                             uvm_va_block_t **new_va_block,
                             uvm_va_range_t *new_va_range);

// Handles a CPU fault in the given VA block, performing any operations
// necessary to establish a coherent CPU mapping (migrations, cache invalidates,
// etc.).
//
// Locking:
//  - vma->vm_mm->mmap_sem must be held in at least read mode. Note, that might
//    not be the same as current->mm->mmap_sem.
//  - va_space lock must be held in at least read mode
//
// Returns NV_ERR_INVALID_ACCESS_TYPE if a CPU mapping to fault_addr cannot be
// accessed, for example because it's within a range group which is non-
// migratable.
NV_STATUS uvm_va_block_cpu_fault(uvm_va_block_t *va_block, NvU64 fault_addr, bool is_write);

// Size of the block in bytes. Guaranteed to be a page-aligned value between
// PAGE_SIZE and UVM_VA_BLOCK_SIZE.
static inline NvU64 uvm_va_block_size(uvm_va_block_t *block)
{
    NvU64 size = block->end - block->start + 1;
    UVM_ASSERT(PAGE_ALIGNED(size));
    UVM_ASSERT(size >= PAGE_SIZE);
    UVM_ASSERT(size <= UVM_VA_BLOCK_SIZE);
    return size;
}

// VA block lookup functions. There are a number of permutations which might be
// useful, such as looking up the block from {va_space, va_range} x {addr,
// block index}. The ones implemented here and in uvm8_va_range.h support the
// primary two use cases, which are:
// 1) Iterating over all VA blocks in a VA range. This uses block indices on the
//    VA range:
//      uvm_va_range_num_blocks
//      uvm_va_range_block_index
//      uvm_va_range_block
//      uvm_va_range_block_create
// 2) Operating on a single VA block (fault). This looks up the block using the
//    VA space and address:
//      uvm_va_block_find
//      uvm_va_block_find_create

// Finds the VA block containing addr, if any. The va_space->lock must be held
// in at least read mode. Return values:
// NV_ERR_INVALID_ADDRESS   addr is not covered by a va_range
//
// NV_ERR_OBJECT_NOT_FOUND  addr is valid but no block has been allocated to
//                          cover it yet
//
// NV_OK                    The block was returned successfully
NV_STATUS uvm_va_block_find(uvm_va_space_t *va_space, NvU64 addr, uvm_va_block_t **out_block);

// Same as uvm_va_block_find except that the block is created if not already
// present in the VA range.
NV_STATUS uvm_va_block_find_create(uvm_va_space_t *va_space, NvU64 addr, uvm_va_block_t **out_block);

NV_STATUS uvm8_test_change_pte_mapping(UVM_TEST_CHANGE_PTE_MAPPING_PARAMS *params, struct file *filp);
NV_STATUS uvm8_test_va_block_info(UVM_TEST_VA_BLOCK_INFO_PARAMS *params, struct file *filp);

#endif // __UVM8_VA_BLOCK_H__
