/*******************************************************************************
    Copyright (c) 2015 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm8_va_space.h"
#include "uvm8_va_range.h"
#include "uvm8_lock.h"
#include "uvm8_global.h"
#include "uvm8_kvmalloc.h"
#include "uvm_common.h"
#include "nv_uvm_interface.h"

static void release_enabled_peers(uvm_va_space_t *va_space, uvm_gpu_t *peer_gpu);
static void remove_gpu_va_space(uvm_va_space_t *va_space, uvm_gpu_t *gpu);

NV_STATUS uvm_va_space_create(struct inode *inode, struct file *filp)
{
    NV_STATUS status;
    uvm_va_space_t *va_space;
    va_space = uvm_kvmalloc_zero(sizeof(*va_space));

    if (!va_space)
        return NV_ERR_NO_MEMORY;

    uvm_init_rwsem(&va_space->lock, UVM_LOCK_ORDER_VA_SPACE);
    uvm_range_tree_init(&va_space->va_range_tree);

    // By default all struct files on the same inode share the same
    // address_space structure (the inode's) across all processes. This means
    // unmap_mapping_range would unmap virtual mappings across all processes on
    // that inode.
    //
    // Since the UVM driver uses the mapping offset as the VA of the file's
    // process, we need to isolate the mappings to each process.
    address_space_init_once(&va_space->mapping);
    va_space->mapping.host = inode;

    // Some paths in the kernel, for example force_page_cache_readahead which
    // can be invoked from user-space via madvise MADV_WILLNEED and fadvise
    // POSIX_FADV_WILLNEED, check the function pointers within
    // file->f_mapping->a_ops for validity. However, those paths assume that a_ops
    // itself is always valid. Handle that by using the inode's a_ops pointer,
    // which is what f_mapping->a_ops would point to anyway if we weren't re-
    // assigning f_mapping.
    va_space->mapping.a_ops = inode->i_mapping->a_ops;

#if defined(NV_ADDRESS_SPACE_HAS_BACKING_DEV_INFO)
    va_space->mapping.backing_dev_info = inode->i_mapping->backing_dev_info;
#endif

    // Init to 0 since we rely on atomic_inc_return behavior to return 1 as the first ID
    atomic64_set(&va_space->range_group_id_counter, 0);

    INIT_RADIX_TREE(&va_space->range_groups, NV_UVM_GFP_FLAGS);

    bitmap_zero(va_space->enabled_peers, UVM8_MAX_UNIQUE_GPU_PAIRS);

    filp->private_data = va_space;
    filp->f_mapping = &va_space->mapping;

    status = uvm_perf_init_va_space_events(va_space, &va_space->perf_events);
    if (status != NV_OK)
        goto fail;

    return NV_OK;

fail:
    uvm_kvfree(va_space);

    return status;
}

void uvm_va_space_destroy(struct file *filp)
{
    uvm_va_space_t *va_space = uvm_va_space_get(filp);
    uvm_va_range_t *va_range, *va_range_next;
    uvm_gpu_t *gpu;

    // Nothing should be able to get to the va space once we are in destroy
    uvm_assert_rwsem_unlocked(&va_space->lock);

    uvm_for_each_va_range_safe(va_range, va_range_next, va_space)
        uvm_va_range_destroy(va_range);

    uvm_range_group_radix_tree_destroy(&va_space->range_groups);

    uvm_down_write(&g_uvm_global.global_lock);

    // Need to take va_space lock just to make sure "remove_gpu_va_space" should not assert.
    uvm_down_write(&va_space->lock);

    for_each_va_space_gpu(gpu, va_space) {
        // For each peer-to-peer pair that was explicitly enabled, but not explicitly disabled, disable it.
        release_enabled_peers(va_space, gpu);

        // If this GPU VA Space was explicitly registered, but not explicitly unregistered,
        // unregister it.
        if (uvm_processor_mask_test(&va_space->registered_gpu_va_spaces, gpu->id))
            remove_gpu_va_space(va_space, gpu);

        // For each GPU that was explicitly registered, but not explicitly unregistered, release it:
        uvm_processor_mask_clear(&va_space->registered_gpus, gpu->id);

        uvm_gpu_release(gpu);
    }

    uvm_perf_destroy_va_space_events(&va_space->perf_events);

    uvm_up_write(&va_space->lock);

    UVM_ASSERT(uvm_processor_mask_empty(&va_space->registered_gpus));
    UVM_ASSERT(uvm_processor_mask_empty(&va_space->registered_gpu_va_spaces));

    uvm_up_write(&g_uvm_global.global_lock);

    uvm_kvfree(va_space);

    filp->private_data = NULL;
    filp->f_mapping = NULL;
}

// Note that the "VA space" in the function name refers to a UVM per-process VA space.
// (This is different from a per-GPU VA space.)
NV_STATUS uvm_va_space_register_gpu(uvm_va_space_t *va_space, NvProcessorUuid *gpu_uuid)
{
    NV_STATUS status;
    uvm_gpu_t *gpu;

    uvm_down_write(&g_uvm_global.global_lock);
    uvm_down_write(&va_space->lock);

    status = uvm_gpu_retain(gpu_uuid, &gpu);
    if (status != NV_OK)
        goto done;

    // Make sure the gpu hasn't been already registered in this va space
    if (uvm_processor_mask_test(&va_space->registered_gpus, gpu->id)) {
        status = NV_ERR_INVALID_DEVICE;
        uvm_gpu_release(gpu);
        goto done;
    }

    uvm_processor_mask_set(&va_space->registered_gpus, gpu->id);

done:
    uvm_up_write(&va_space->lock);
    uvm_up_write(&g_uvm_global.global_lock);

    return status;
}

NV_STATUS uvm_va_space_unregister_gpu(uvm_va_space_t *va_space, NvProcessorUuid *gpu_uuid)
{
    NV_STATUS status = NV_OK;
    uvm_gpu_t *gpu;
    uvm_va_range_t *va_range;

    uvm_down_write(&g_uvm_global.global_lock);
    uvm_down_write(&va_space->lock);

    gpu = uvm_gpu_get_registered_by_uuid(va_space, gpu_uuid);
    if (!gpu) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    // If this GPU has any peer-to-peer pair that was explicitly enabled,
    // but not explicitly disabled, disable it.
    release_enabled_peers(va_space, gpu);

    // If a GPU VA Space was explicitly registered, but not explicitly unregistered,
    // unregister it.
    if (uvm_processor_mask_test(&va_space->registered_gpu_va_spaces, gpu->id))
        remove_gpu_va_space(va_space, gpu);

    // Reset preferred location and accessed by of VA ranges if needed
    uvm_for_each_va_range(va_range, va_space) {
        if (va_range->preferred_location == gpu->id) {
            status = uvm_va_range_set_preferred_location(va_range, UVM8_MAX_PROCESSORS);
            if (status != NV_OK)
                goto done;
        }

        uvm_processor_mask_clear(&va_range->accessed_by, gpu->id);
    }

    uvm_processor_mask_clear(&va_space->registered_gpus, gpu->id);
    uvm_gpu_release(gpu);

done:
    uvm_up_write(&va_space->lock);
    uvm_up_write(&g_uvm_global.global_lock);

    return status;
}

static void release_enabled_peers(uvm_va_space_t *va_space, uvm_gpu_t *peer_gpu)
{
    NvU32 table_index = 0;
    uvm_gpu_t *gpu;

    UVM_ASSERT(va_space);
    UVM_ASSERT(peer_gpu);

    for_each_gpu_in_mask(gpu, &va_space->registered_gpus) {
        if (gpu->id != peer_gpu->id) {
            table_index = uvm_gpu_peer_table_index(gpu->id, peer_gpu->id);
            if (test_bit(table_index, va_space->enabled_peers)) {
                uvm_gpu_release_peer_access(gpu, peer_gpu);
                __clear_bit(table_index, va_space->enabled_peers);
            }
        }
    }
}

NV_STATUS uvm_va_space_enable_peer_access(uvm_va_space_t *va_space,
                                          NvProcessorUuid *gpu_uuid_1,
                                          NvProcessorUuid *gpu_uuid_2)
{
    NV_STATUS status = NV_OK;
    uvm_gpu_t *gpu[2] = { 0 };
    NvU32 table_index = 0;

    uvm_down_write(&g_uvm_global.global_lock);
    uvm_down_write(&va_space->lock);

    gpu[0] = uvm_gpu_get_registered_by_uuid(va_space, gpu_uuid_1);
    gpu[1] = uvm_gpu_get_registered_by_uuid(va_space, gpu_uuid_2);

    if (!gpu[0] || !gpu[1]) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    if (gpu[0]->id == gpu[1]->id) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    table_index = uvm_gpu_peer_table_index(gpu[0]->id, gpu[1]->id);

    // If this peer to peer pair is already enabled for a given VA, error out.
    if (test_bit(table_index, va_space->enabled_peers)) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    status = uvm_gpu_retain_peer_access(gpu[0], gpu[1]);
    if (status != NV_OK)
        goto done;

    __set_bit(table_index, va_space->enabled_peers);

done:
    uvm_up_write(&va_space->lock);
    uvm_up_write(&g_uvm_global.global_lock);
    return status;
}

NV_STATUS uvm_va_space_disable_peer_access(uvm_va_space_t *va_space,
                                           NvProcessorUuid *gpu_uuid_1,
                                           NvProcessorUuid *gpu_uuid_2)
{
    NV_STATUS status = NV_OK;
    uvm_gpu_t *gpu[2] = { 0 };
    NvU32 table_index = 0;

    uvm_down_write(&g_uvm_global.global_lock);
    uvm_down_write(&va_space->lock);

    gpu[0] = uvm_gpu_get_registered_by_uuid(va_space, gpu_uuid_1);
    gpu[1] = uvm_gpu_get_registered_by_uuid(va_space, gpu_uuid_2);

    if (!gpu[0] || !gpu[1]) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    if (gpu[0]->id == gpu[1]->id) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    table_index = uvm_gpu_peer_table_index(gpu[0]->id, gpu[1]->id);

    // If this peer to peer pair is not enabled for a given VA, error out.
    if (!test_bit(table_index, va_space->enabled_peers)) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    uvm_gpu_release_peer_access(gpu[0], gpu[1]);

    __clear_bit(table_index, va_space->enabled_peers);

done:
    uvm_up_write(&va_space->lock);
    uvm_up_write(&g_uvm_global.global_lock);
    return status;
}


static NV_STATUS add_gpu_va_space(uvm_gpu_t *gpu,
                                  uvm_va_space_t *va_space,
                                  NvHandle client_handle,
                                  NvHandle gpu_va_space_handle)
{
    NV_STATUS status = NV_OK;
    uvm_gpu_va_space_t *gpu_va_space;
    NvU32 big_page_size = 0;

    UVM_ASSERT(gpu);
    UVM_ASSERT(va_space);
    uvm_assert_rwsem_locked_write(&va_space->lock);

    gpu_va_space = &va_space->gpu_va_spaces[gpu->id - 1];

    // TODO:- This interface needs to use controlFd to do validation checks.
    status = uvm_rm_locked_call(nvUvmInterfaceDupAddressSpace(g_uvm_global.rm_session_handle,
                                                              gpu->uuid.uuid,
                                                              client_handle,
                                                              gpu_va_space_handle,
                                                              &gpu_va_space->duped_gpu_va_space));
    if (status != NV_OK) {
        UVM_DBG_PRINT("failed to dup address space with error: %s, for GPU:%s \n",
                nvstatusToString(status), gpu->name);
        goto cleanup;
    }

    status = uvm_rm_locked_call(nvUvmInterfaceGetBigPageSize(gpu_va_space->duped_gpu_va_space, &big_page_size));
    if (status != NV_OK) {
        UVM_DBG_PRINT("failed to get a big_page_size with error: %s, for GPU:%s \n",
                nvstatusToString(status), gpu->name);
        goto cleanup;
    }

    UVM_ASSERT(big_page_size);

    // TODO:-
    // 1. Allocate a new root PDB.
    // 2. Replace existing root PDB associated with the GPU VA space with the new root PDB.

    uvm_processor_mask_set(&va_space->registered_gpu_va_spaces, gpu->id);

    return NV_OK;

cleanup:
    remove_gpu_va_space(va_space, gpu);
    return status;
}

NV_STATUS uvm_va_space_register_gpu_va_space(uvm_va_space_t *va_space,
                                             NvHandle client_handle,
                                             NvHandle gpu_va_space_handle,
                                             NvProcessorUuid *gpu_uuid)
{
    NV_STATUS status = NV_OK;
    uvm_gpu_t *gpu = NULL;

    uvm_down_read(&g_uvm_global.global_lock);
    uvm_down_write(&va_space->lock);

    gpu = uvm_gpu_get_registered_by_uuid(va_space, gpu_uuid);
    if (!gpu) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    // Error out if the GPU VA space is already registered.
    if (uvm_processor_mask_test(&va_space->registered_gpu_va_spaces, gpu->id)) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    status = add_gpu_va_space(gpu, va_space, client_handle, gpu_va_space_handle);

done:
    uvm_up_write(&va_space->lock);
    uvm_up_read(&g_uvm_global.global_lock);
    return status;
}

static void remove_gpu_va_space(uvm_va_space_t *va_space, uvm_gpu_t *gpu)
{
    uvm_gpu_va_space_t *gpu_va_space;

    UVM_ASSERT(gpu);
    UVM_ASSERT(va_space);
    uvm_assert_rwsem_locked_write(&va_space->lock);

    gpu_va_space = &va_space->gpu_va_spaces[gpu->id - 1];

    // TODO:-
    // 1. Unmap all page tables in this VA space on this GPU.
    // 2. Unregister all channels for this GPU + VA space.
    // 3. Restore the original root PDB with GPU VA.
    // 4. Free the newly allocated root PDB.

    if (gpu_va_space->duped_gpu_va_space)
        uvm_rm_locked_call_void(nvUvmInterfaceAddressSpaceDestroy(gpu_va_space->duped_gpu_va_space));

    memset(gpu_va_space, 0, sizeof(uvm_gpu_va_space_t));

    uvm_processor_mask_clear(&va_space->registered_gpu_va_spaces, gpu->id);
}

NV_STATUS uvm_va_space_unregister_gpu_va_space(uvm_va_space_t *va_space, NvProcessorUuid *gpu_uuid)
{
    NV_STATUS status = NV_OK;
    uvm_gpu_t *gpu = NULL;

    UVM_ASSERT(va_space);

    uvm_down_read(&g_uvm_global.global_lock);
    uvm_down_write(&va_space->lock);

    gpu = uvm_gpu_get_registered_by_uuid(va_space, gpu_uuid);

    if (!gpu) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    // If the GPU VA space is not registered, error out.
    if (!uvm_processor_mask_test(&va_space->registered_gpu_va_spaces, gpu->id)) {
        status = NV_ERR_INVALID_DEVICE;
        goto done;
    }

    remove_gpu_va_space(va_space, gpu);

done:
    uvm_up_write(&va_space->lock);
    uvm_up_read(&g_uvm_global.global_lock);
    return status;
}
